/*****************************************************************************
 * gyach.c
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * Copyright (C) 2000 Chris Pinkham
 * Released under the terms of the GPL.
 * *NO WARRANTY*
 *
 * cpinkham@infi.net
 * http://www4.infi.net/~cpinkham/gyach/
 *****************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <glib.h>

#include "config.h"

#ifdef USE_PTHREAD_CREATE
#include <pthread.h>
#endif

#include "parsecfg.h"
#include "gyach.h"
#include "images.h"
#include "yahoolib.h"
#include "commands.h"
#include "callbacks.h"

#include "interface.h"
#include "support.h"
#include "nix.h"
#include "main.h"

int debug_packets = 0;

GtkWidget * login_window;
GtkWidget * chat_window;
GtkWidget * chat_text;
GtkWidget * chat_status;
GtkWidget * chat_entry;
GtkWidget * chat_emotions;
GtkWidget * chat_users;
GtkWidget * chat_user_menu;
GtkWidget * setup_menu;
GtkWidget * emote_menu;
GtkWidget * emote_window;
GtkWidget * pm_window;
GtkWidget * pm_entry;
GtkWidget * pm_user;
GtkWidget * chat_username;
GtkWidget * chat_password;
GtkWidget * chat_roomname;
GtkWidget * profile_window = NULL;
GtkWidget * about_box = NULL;
GtkWidget * setup_window = NULL;
GtkWidget * room_window = NULL;

GtkStyle  *profile_style = NULL;

guint st_cid;	/* statusbar context_id */

struct yahoo_context yahoo_ctx;

int  emotion_count;
char *user_selected = NULL;
char *room_selected = NULL;
int  emote_selected = -1;
char *follow_user = NULL;
char *capture_filename = NULL;
FILE *capture_fp = NULL;

int child_pid = 0;
int my_status = 0;
time_t auto_away_time;
time_t child_timeout;

char *prof_text = NULL;
char *prof_pic = NULL;

int  io_callback_tag = -1;
int  ping_callback_tag = -1;
int  logged_in = 0;
int  show_blended_colors = 1;
int  show_colors = 1;
int  show_fonts = 0;
int  def_user = -1;
int  auto_away = 15;
int  auto_login = 0;
int  pm_brings_back = 0;
int  buffer_size = 20;
int  show_html = 1;
int  show_statuses = 1;
int  show_enters = 1;
int  profile_viewer = 0;
int  ignore_on_mults = 0;
int  use_bold = 0;
int  use_underline = 0;
int  use_italics = 0;
int  font_size = 14;
int  force_lowercase = 0;
int  ping_minutes = 10;
int  suppress_mult = 0;
int  chat_window_x = 0;
int  chat_window_y = 0;
int  chat_window_width = 630;
int  chat_window_height = 450;
int  chat_text_width = 480;
int  chat_user_height = 200;
int  disp_auto_ignored = 1;
int  remember_position = 0;
int  remember_password = 1;
int  limit_lfs = 15;
int  auto_reply_when_away = 0;
int  pm_in_sep_windows = 1;
int  auto_raise_pm = 1;
int  auto_close_roomlist = 1;
int  chat_timestamp = 0;
int  chat_timestamp_pm = 0;
int  ignore_guests = 0;
int  auto_reconnect = 1;
char *auto_reply_msg = NULL;
char *browser_command = NULL;
char *filter_command = NULL;
char *font_name = NULL;
char *font_family = NULL;
char *use_color = NULL;
char *custom_away_message = NULL;
char *username0 = NULL;
char *username1 = NULL;
char *username2 = NULL;
char *username3 = NULL;
char *username4 = NULL;
char *password = NULL;
char *def_room  = NULL;
char *fav_room0 = NULL;
char *fav_room1 = NULL;
char *fav_room2 = NULL;
char *fav_room3 = NULL;
char *fav_room4 = NULL;
char *fav_room5 = NULL;
char *fav_room6 = NULL;
char *fav_room7 = NULL;
char *fav_room8 = NULL;
char *fav_room9 = NULL;

char cfg_filename[ 1024 ] = "";
char cfg_dir[1024] = "";

char username[41] = "";
char roomname[129] = "";

int connected = 0;
int valid_user = 0;

GList *ignore_list = NULL;
GList *mute_list = NULL;
GList *regex_list = NULL;
GList *pm_list = NULL;
GList *favroom_list = NULL;
GList *login_list = NULL;
GList *friend_list = NULL;
GList *ofriend_list = NULL;

cfgList *username_list = NULL;
cfgList *fav_room_list = NULL;

int history_limit = 50;
int history_size = 0;
int history_current = -1;

GList *history_list = NULL;

SESSION *GlobalSessionHook;

void append_to_textbox(GtkWidget * window, GtkWidget * textbox, char *text);

cfgStruct cfg[] = {
	/* parameter  type address of variable */
	{ "username",			CFG_STRING_LIST,	&username_list },
	{ "fav_room",			CFG_STRING_LIST,	&fav_room_list },
	{ "show_blended_colors",	CFG_INT,		&show_blended_colors },
	{ "show_colors",			CFG_INT,		&show_colors },
	{ "show_fonts",				CFG_INT,		&show_fonts },
	{ "debug_packets",			CFG_INT,		&debug_packets },
	{ "def_user",				CFG_INT,		&def_user },
	{ "auto_away",				CFG_INT,		&auto_away },
	{ "auto_login",				CFG_INT,		&auto_login },
	{ "pm_brings_back",			CFG_INT,		&pm_brings_back },
	{ "buffer_size",			CFG_INT,		&buffer_size },
	{ "show_html",				CFG_INT,		&show_html },
	{ "show_statuses",			CFG_INT,		&show_statuses },
	{ "show_enters",			CFG_INT,		&show_enters },
	{ "profile_viewer",			CFG_INT,		&profile_viewer },
	{ "ignore_on_mults",		CFG_INT,		&ignore_on_mults },
	{ "use_bold",				CFG_INT,		&use_bold },
	{ "use_italics",			CFG_INT,		&use_italics },
	{ "use_underline",			CFG_INT,		&use_underline },
	{ "force_lowercase",		CFG_INT,		&force_lowercase },
	{ "suppress_mult",			CFG_INT,		&suppress_mult },
	{ "ping_minutes",			CFG_INT,		&ping_minutes },
	{ "chat_window_x",			CFG_INT,		&chat_window_x },
	{ "chat_window_y",			CFG_INT,		&chat_window_y },
	{ "chat_window_width",		CFG_INT,		&chat_window_width },
	{ "chat_window_height",		CFG_INT,		&chat_window_height },
	{ "chat_text_width",		CFG_INT,		&chat_text_width },
	{ "chat_user_height",		CFG_INT,		&chat_user_height },
	{ "disp_auto_ignored",		CFG_INT,		&disp_auto_ignored },
	{ "remember_position",		CFG_INT,		&remember_position },
	{ "remember_password",		CFG_INT,		&remember_password },
	{ "limit_lfs",				CFG_INT,		&limit_lfs },
	{ "auto_reply_when_away",	CFG_INT,		&auto_reply_when_away },
	{ "pm_in_sep_windows",		CFG_INT,		&pm_in_sep_windows },
	{ "auto_raise_pm",			CFG_INT,		&auto_raise_pm },
	{ "auto_close_roomlist",	CFG_INT,		&auto_close_roomlist },
	{ "chat_timestamp",			CFG_INT,		&chat_timestamp },
	{ "chat_timestamp_pm",		CFG_INT,		&chat_timestamp_pm },
	{ "ignore_guests",			CFG_INT,		&ignore_guests },
	{ "auto_reconnect",			CFG_INT,		&auto_reconnect },
	{ "history_limit",			CFG_INT,		&history_limit },
	{ "auto_reply_msg",			CFG_STRING,		&auto_reply_msg },
	{ "use_color",				CFG_STRING,		&use_color },
	{ "custom_away_message",	CFG_STRING,		&custom_away_message },
	{ "browser_command",		CFG_STRING,		&browser_command },
	{ "filter_command",			CFG_STRING,		&filter_command },
	{ "font_name",				CFG_STRING,		&font_name },
	{ "font_family",			CFG_STRING,		&font_family },
	{ "font_size",				CFG_INT,		&font_size },
	{ "username0",				CFG_STRING,		&username0 },
	{ "username1",				CFG_STRING,		&username1 },
	{ "username2",				CFG_STRING,		&username2 },
	{ "username3",				CFG_STRING,		&username3 },
	{ "username4",				CFG_STRING,		&username4 },
	{ "password",				CFG_STRING,		&password },
	{ "def_room",				CFG_STRING,		&def_room },
	{ "fav_room0",				CFG_STRING,		&fav_room0 },
	{ "fav_room1",				CFG_STRING,		&fav_room1 },
	{ "fav_room2",				CFG_STRING,		&fav_room2 },
	{ "fav_room3",				CFG_STRING,		&fav_room3 },
	{ "fav_room4",				CFG_STRING,		&fav_room4 },
	{ "fav_room5",				CFG_STRING,		&fav_room5 },
	{ "fav_room6",				CFG_STRING,		&fav_room6 },
	{ "fav_room7",				CFG_STRING,		&fav_room7 },
	{ "fav_room8",				CFG_STRING,		&fav_room8 },
	{ "fav_room9",				CFG_STRING,		&fav_room9 },
	{ NULL,						CFG_END,		NULL }
};

/* fixme */
void no_home_dir_message() {
	DBG( 1, "no_home_dir_message()\n" );
	/* print out a message or open a window or something */
}

gint gstrcmp( gpointer a, gpointer b ) {
	if ( a && b ) {
		DBG( 31, "gstrcmp( '%s', '%s' )\n", (char *)a, (char *)b );
	} else if ( a ) {
		DBG( 31, "gstrcmp( '%s', NULL )\n", (char *)a );
	} else if ( b ) {
		DBG( 31, "gstrcmp( NULL, '%s' )\n", (char *)b );
	} else {
		DBG( 31, "gstrcmp( NULL, NULL )\n" );
	}

	if (( ! a ) &&
		( ! b )) {
		DBG( 32, "RETURN gstrcmp() == 0\n" );
		return( 0 );
	}

	if ( ! a ) {
		DBG( 32, "RETURN gstrcmp() == -1\n" );
		return( -1 );
	}

	if ( ! b ) {
		DBG( 32, "RETURN gstrcmp() == 1\n" );
		return( 1 );
	}

	DBG( 32, "RETURN gstrcmp() == %d\n", strcasecmp( (char *)a, (char *)b ));

	return( strcasecmp( (char *)a, (char *)b ));
}

void gyach_init() {
	DBG( 1, "gyach_init()\n" );

	read_config();
	ignore_load();

	history_list = g_list_append( history_list, strdup( "" ));
	history_size = 1;

	auto_away_time = time(NULL) + ( auto_away * 60 );

	if ( ! font_name )
		font_name = strdup(
			"-*-times-medium-r-normal-*-14-*-*-*-*-*-*-*" );

	if ( ! font_family )
		font_family = strdup( "times" );

	if ( font_size < 0 )
		font_size = 14;

	if ( ! browser_command ) {
		browser_command =
			strdup( "netscape -noraise -remote openURL\\(%s,Gyach\\)" );
	}

	memset( &yahoo_ctx, 0, sizeof( yahoo_ctx ));
}

int yahoo_ctx_init( void ) {
	/* setup the yahoolib context structure */
	yahoo_ctx.user = sess->user;
	yahoo_ctx.password = sess->pass;
	yahoo_ctx.connect_mode = YAHOO_CONNECT_NORMAL;

	if ( yahoo_ctx.cookie )
		free( yahoo_ctx.cookie );
	
	yahoo_ctx.cookie = strdup( sess->cook );

	if ( yahoo_ctx.login_cookie )
		free( yahoo_ctx.login_cookie );

	yahoo_ctx.login_cookie = strdup( sess->cook );

	if ( yahoo_ctx.login_id )
		free( yahoo_ctx.login_id );

	yahoo_ctx.login_id = strdup( sess->user );

	return( 0 );
}

void process_gtk_events() {
	DBG( 251, "process_gtk_events()\n" );

	return;

	while( gtk_events_pending())
		gtk_main_iteration();
}

void set_config() {
	GtkWidget *tmp_widget;

	DBG( 1, "set_config()\n" );

	tmp_widget = lookup_widget( chat_window, "tgl_bold" );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(tmp_widget), use_bold );

	tmp_widget = lookup_widget( chat_window, "tgl_italics" );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(tmp_widget), use_italics );

	tmp_widget = lookup_widget( chat_window, "tgl_underline" );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(tmp_widget),
		use_underline );
}

void read_config() {
	struct stat sbuf;
	cfgList *this_item;

	DBG( 1, "read_config()\n" );

	if ( ! getenv( "HOME" )) {
		no_home_dir_message();
		return;
	}

	if ( cfg_filename[ 0 ] == '\0' )
		sprintf( cfg_filename, "%s/.gyach/gyachrc", getenv( "HOME" ));

	if ( stat( cfg_filename, &sbuf ))
		write_config();

	cfgParse( cfg_filename, cfg, CFG_SIMPLE );

	if ( ! use_color )
		use_color = strdup( "black" );

	if ( username0 )
		login_list = g_list_append( login_list, strdup( username0 ));
	if ( username1 )
		login_list = g_list_append( login_list, strdup( username1 ));
	if ( username2 )
		login_list = g_list_append( login_list, strdup( username2 ));
	if ( username3 )
		login_list = g_list_append( login_list, strdup( username3 ));
	if ( username4 )
		login_list = g_list_append( login_list, strdup( username4 ));

	this_item = username_list;
	while( this_item ) {
		login_list = g_list_append( login_list, strdup( this_item->str ));
		this_item = this_item->next;
	}

	if (( fav_room0 ) && ( fav_room0[0] ))
		favroom_list = g_list_append( favroom_list, strdup( fav_room0 ));
	if (( fav_room1 ) && ( fav_room1[0] ))
		favroom_list = g_list_append( favroom_list, strdup( fav_room1 ));
	if (( fav_room2 ) && ( fav_room2[0] ))
		favroom_list = g_list_append( favroom_list, strdup( fav_room2 ));
	if (( fav_room3 ) && ( fav_room3[0] ))
		favroom_list = g_list_append( favroom_list, strdup( fav_room3 ));
	if (( fav_room4 ) && ( fav_room4[0] ))
		favroom_list = g_list_append( favroom_list, strdup( fav_room4 ));
	if (( fav_room5 ) && ( fav_room5[0] ))
		favroom_list = g_list_append( favroom_list, strdup( fav_room5 ));
	if (( fav_room6 ) && ( fav_room6[0] ))
		favroom_list = g_list_append( favroom_list, strdup( fav_room6 ));
	if (( fav_room7 ) && ( fav_room7[0] ))
		favroom_list = g_list_append( favroom_list, strdup( fav_room7 ));
	if (( fav_room8 ) && ( fav_room8[0] ))
		favroom_list = g_list_append( favroom_list, strdup( fav_room8 ));
	if (( fav_room9 ) && ( fav_room9[0] ))
		favroom_list = g_list_append( favroom_list, strdup( fav_room9 ));

	this_item = fav_room_list;
	while( this_item ) {
		favroom_list = g_list_append( favroom_list, strdup( this_item->str ));
		this_item = this_item->next;
	}
}

void write_config() {
	GtkWidget *tmp_widget;
	GList *this_item = NULL;
	FILE *gyachrc;
	struct stat sbuf;

	DBG( 1, "write_config()\n" );

	if ( ! getenv( "HOME" )) {
		no_home_dir_message();
		return;
	}

	if ( cfg_dir[ 0 ] == '\0' )
		sprintf( cfg_dir, "%s/.gyach", getenv( "HOME" ));

	if ( stat( cfg_dir, &sbuf ))
		mkdir( cfg_dir, 0700 );

	if ( cfg_filename[ 0 ] == '\0' )
		sprintf( cfg_filename, "%s/.gyach/gyachrc", getenv( "HOME" ));

	gyachrc = fopen( cfg_filename, "w" );

	if ( gyachrc ) {
		if ( chat_window ) {
			tmp_widget = chat_window;

			if ( remember_position ) {
				gdk_window_get_position( tmp_widget->window,
					&chat_window_x, &chat_window_y );
			}

			/* get current sizes */
			chat_window_width = tmp_widget->allocation.width;
			chat_window_height = tmp_widget->allocation.height;

			tmp_widget = lookup_widget( chat_window, "chat_text" );
			chat_text_width = tmp_widget->allocation.width;

			tmp_widget = lookup_widget( chat_window, "scrolledwindow3" );
			chat_user_height = tmp_widget->allocation.height;
		}

		/* FIXME, put while loop here to save "username" */
		this_item = login_list;
		while( this_item ) {
			fprintf( gyachrc, "username = \"%s\"\n", (char *)this_item->data );
			this_item = g_list_next( this_item );
		}

		if (password)  fprintf( gyachrc, "password = \"%s\"\n", password );

		this_item = favroom_list;
		while( this_item ) {
			fprintf( gyachrc, "fav_room = \"%s\"\n", (char *)this_item->data );
			this_item = g_list_next( this_item );
		}

		if (auto_reply_msg) fprintf( gyachrc,
			"auto_reply_msg = \"%s\"\n", auto_reply_msg );
		if (custom_away_message) fprintf( gyachrc,
			"custom_away_message = \"%s\"\n", custom_away_message );
		if (use_color) fprintf( gyachrc,
			"use_color = \"%s\"\n", use_color );
		if (font_name) fprintf( gyachrc,
			"font_name = \"%s\"\n", font_name );
		if (font_family) fprintf( gyachrc,
			"font_family = \"%s\"\n", font_family );
		if (browser_command) fprintf( gyachrc,
			"browser_command = \"%s\"\n", browser_command );
		if (filter_command) fprintf( gyachrc,
			"filter_command = \"%s\"\n", filter_command );

		fprintf( gyachrc, "show_blended_colors = %d\n", show_blended_colors );
		fprintf( gyachrc, "show_colors = %d\n", show_colors );
		fprintf( gyachrc, "show_fonts = %d\n", show_fonts );
		fprintf( gyachrc, "debug_packets = %d\n", debug_packets );
		fprintf( gyachrc, "auto_away = %d\n", auto_away );
		fprintf( gyachrc, "auto_login = %d\n", auto_login );
		fprintf( gyachrc, "pm_brings_back = %d\n", pm_brings_back );
		fprintf( gyachrc, "buffer_size = %d\n", buffer_size );
		fprintf( gyachrc, "show_html = %d\n", show_html );
		fprintf( gyachrc, "show_statuses = %d\n", show_statuses );
		fprintf( gyachrc, "show_enters = %d\n", show_enters );
		fprintf( gyachrc, "profile_viewer = %d\n", profile_viewer );
		fprintf( gyachrc, "ignore_on_mults = %d\n", ignore_on_mults );
		fprintf( gyachrc, "use_bold = %d\n", use_bold );
		fprintf( gyachrc, "use_italics = %d\n", use_italics );
		fprintf( gyachrc, "use_underline = %d\n", use_underline );
		fprintf( gyachrc, "force_lowercase = %d\n", force_lowercase );
		fprintf( gyachrc, "suppress_mult = %d\n", suppress_mult );
		fprintf( gyachrc, "ping_minutes = %d\n", ping_minutes );
		fprintf( gyachrc, "font_size = %d\n", font_size );

		if ( remember_position ) {
			fprintf( gyachrc, "chat_window_x = %d\n", chat_window_x );
			fprintf( gyachrc, "chat_window_y = %d\n", chat_window_y );
		}

		fprintf( gyachrc, "chat_window_width = %d\n", chat_window_width );
		fprintf( gyachrc, "chat_window_height = %d\n", chat_window_height );
		fprintf( gyachrc, "chat_text_width = %d\n", chat_text_width );
		fprintf( gyachrc, "chat_user_height = %d\n", chat_user_height );
		fprintf( gyachrc, "disp_auto_ignored = %d\n", disp_auto_ignored );
		fprintf( gyachrc, "remember_position = %d\n", remember_position );
		fprintf( gyachrc, "remember_password = %d\n", remember_password );
		fprintf( gyachrc, "limit_lfs = %d\n", limit_lfs );
		fprintf( gyachrc, "auto_reply_when_away = %d\n", auto_reply_when_away );
		fprintf( gyachrc, "pm_in_sep_windows = %d\n", pm_in_sep_windows );
		fprintf( gyachrc, "auto_raise_pm = %d\n", auto_raise_pm );
		fprintf( gyachrc, "auto_close_roomlist = %d\n", auto_close_roomlist );
		fprintf( gyachrc, "chat_timestamp = %d\n", chat_timestamp );
		fprintf( gyachrc, "chat_timestamp_pm = %d\n", chat_timestamp_pm );
		fprintf( gyachrc, "ignore_guests = %d\n", ignore_guests );
		fprintf( gyachrc, "auto_reconnect = %d\n", auto_reconnect );
		fprintf( gyachrc, "history_limit = %d\n", history_limit );

		fclose( gyachrc );
	}
}

/*********************************/
/* Default emotions/aliases list */
/*********************************/
struct EMOTION {
	char *em_name;
	char *em_everyone;
	char *em_user;
};

static struct EMOTION emotions[] = {
	{ "Agree",	"agrees whole-heartedly.", "agrees with $0." },
	{ "Apologize",	"apologizes.", "apologizes to $0." },
	{ "Applaud",	"applauds fervently.", "applauds $0 fervently hmmm." },
	{ "Beam",	"beams.", "beams at $0." },
	{ "Beg",	"begs like a dog.", "begs $0." },
	{ "Blush",	"blushes in embarassment.", "blushes at $0 in embarassment." },
	{ "Bow",	"bows gracefully.", "bows to $0 gracefully." },
	{ "Cheer",	"cheers enthusiastically.", "cheers $0 enthusiastically." },
	{ "Clap",	"claps briefly.", "claps for $0 briefly." },
	{ "Disagree",	"disagrees.", "disagrees with $0." },
	{ "Faint",	"faints dead away.", "faints dead away at the sight of $0's." },
	{ "Flirt",	"flirts.", "flirts with $0." },
	{ "Grin",	"grins evilly.", "grins at $0 evilly." },
	{ "Grovel",	"grovels.", "grovels at $0's feet." },
	{ "Hug",	"waves Hi to everyone.", "gives $0 a great big bear hug." },
	{ "Kiss",	"kisses.", "kisses $0." },
	{ "Laugh",	"laughs hysterically.", "laughs at $0." },
	{ "Listen",	"listens.", "listens to $0 hanging on her every word." },
	{ "Melt",	"melts.", "melts in $0's arms." },
	{ "Misgrin",	"grins mischievously.", "grins at $0 mischievously." },
	{ "Nod",	"nods solemnly.", "nods to $0 approvingly." },
	{ "Nudge",	"nudges everyone in the side.", "nudges $0 in the side." },
	{ "Pout",	"pouts.", "pouts at $0." },
	{ "Sit",	"sits down.", "sits down next to $0." },
	{ "Smile",	"smiles.", "smiles at $0." },
	{ "Stand",	"stands up.", "stands up next to $0." },
	{ "Wave",	"waves at everyone.", "waves at $0." },
	{ NULL, NULL, NULL }
};

/*
 * emotion datafile fmt is "emotion]everyone|individual" similar to
 * CheetaChat but support's multiple arguments per emote ($0, $1, ... $9 )
 */
void read_emotions() {
	gchar *listitem[3];
	FILE *fp;
	struct EMOTION *em;
	char line[1024];
	char *emotion = NULL;
	char *em_everyone = NULL;
	char *em_user = NULL;
	char *ptr = NULL;
	char emot_filename[1024] = "";
	int row;

	DBG( 1, "read_emotions()\n" );

	gtk_clist_set_sort_column(GTK_CLIST(chat_emotions), 0 );
	gtk_clist_set_auto_sort(GTK_CLIST(chat_emotions), 1 );

	gtk_clist_freeze(GTK_CLIST(chat_emotions));

	/* remove any items already in the list */
	while(1) {
		if ( !gtk_clist_get_text(GTK_CLIST(chat_emotions), 0, 0, &emotion )) {
			break;
		}
		gtk_clist_remove( GTK_CLIST(chat_emotions), 0 );
	}

	listitem[0] = emotion;
	listitem[1] = em_everyone;
	listitem[2] = em_user;
	emotion = line;

	if ( emot_filename[ 0 ] == '\0' )
		sprintf( emot_filename, "%s/.gyach/emotions.dat",
			getenv( "HOME" ));

	fp = fopen( emot_filename, "r" );

	emotion_count = 0;

	if ( fp ) {
		while( fgets( line, 1023, fp )) {
			if ( ! strchr( line, ']' ))	continue;

			while(( line[strlen(line)-1] == '\r' ) ||
				  ( line[strlen(line)-1] == '\n' )) {
				line[strlen(line)-1] = '\0';
			}

			while(( ptr = strstr( line, "%N" )) != NULL ) {
				*ptr = '$';
				*(ptr+1) = '0';
			}

			emotion_count++;

			ptr = line;
			while( *ptr != ']' ) {
				*ptr = tolower( *ptr );
				ptr++;
			}

			line[0] = toupper( line[0] );
			ptr = strchr( line, ']' );
			*ptr = '\0';

			em_everyone = ptr + 1;
			ptr++;

			ptr = strchr( em_everyone, '|' );
			*ptr = '\0';
			em_user = ptr + 1;

			listitem[0] = emotion;
			row = gtk_clist_insert(GTK_CLIST(chat_emotions), 0, listitem );
			gtk_clist_set_text(GTK_CLIST(chat_emotions), row, 1, em_everyone );
			gtk_clist_set_text(GTK_CLIST(chat_emotions), row, 2, em_user );
		}
		fclose( fp );
	} else {
		/* no emotion file so set defaults */
		em = &emotions[0];
		while( em->em_name ) {
			listitem[0] = em->em_name;
			row = gtk_clist_insert(GTK_CLIST(chat_emotions), 0, listitem );
			gtk_clist_set_text(GTK_CLIST(chat_emotions), row, 1,
				em->em_everyone );
			gtk_clist_set_text(GTK_CLIST(chat_emotions), row, 2,
				em->em_user );
			
			em++;
		}

		/* save our default set to the file */
		save_emotions();
	}

	gtk_clist_thaw(GTK_CLIST(chat_emotions));
}

void save_emotions() {
	gchar *em_name;
	gchar *em_everyone;
	gchar *em_user;
	char emot_filename[1024] = "";
	int i = 0;
	FILE *fp;

	DBG( 1, "save_emotions()\n" );

	if ( emot_filename[ 0 ] == '\0' )
		sprintf( emot_filename, "%s/.gyach/emotions.dat",
			getenv( "HOME" ));

	fp = fopen( emot_filename, "wb" );

	if ( fp ) {
		while(1) {
			if ( !gtk_clist_get_text(GTK_CLIST(chat_emotions), i, 0,&em_name)) {
				break;
			}

			gtk_clist_get_text(GTK_CLIST(chat_emotions), i, 1, &em_everyone );
			gtk_clist_get_text(GTK_CLIST(chat_emotions), i, 2, &em_user );

			fprintf( fp, "%s]%s|%s\n", em_name, em_everyone, em_user );
			
			i++;
		}
		fclose( fp );
	}
}

int alias_find( char *alias ) {
	gchar *emotion;
	gchar tmp_em[21];
	int i;
	int found = -1;

	DBG( 11, "alias_find( '%s' )\n", alias );

	i = 0;
	while( 1 ) {
		if ( !gtk_clist_get_text(GTK_CLIST(chat_emotions), i, 0, &emotion ))
			break;

		strcpy( tmp_em, emotion );

		if ( ! strcasecmp( alias, tmp_em )) {
			found = i;
			break;
		}

		i++;
	}

	DBG( 12, "RETURN alias_find() == %d\n", found );

	return( found );
}

void strip_html_tags( char *str ) {
	char *to, *from;

	DBG( 21, "strip_html_tags()\n" );

	to = str;
	from = str;

	while( *from ) {
		if ( *from == '<' ) {
			from++;
			while(( *from != '>' ) &&
				  ( *from != '<' )) {
				from++;
			}
			from++;
		} else {
			*to = *from;
			to++;
			from++;
		}
	}
	*to = *from;

	to = str;
	from = str;
	while( *from ) {
		if ( *from == '\r' ) {
			from++;
		} else if ( *from == '&' ) {
			if ( *(from+1) == '#' ) {
				while( *from != ';' ) {
					from++;
				}
			} else if (( *(from+1) == 'n' ) &&
					   ( *(from+2) == 'b' ) &&
					   ( *(from+3) == 's' ) &&
					   ( *(from+4) == 'p' ) &&
					   ( *(from+5) == ';' )) {
				from += 5;
				*to = ' ';
				to++;
			}
			from++;
		} else {
			*to = *from;
			to++;
			from++;
		}
	}
	*to = *from;

	to = str;
	from = str;
	while( *from ) {
		if (( *from == '\n' ) &&
			( *(from+1) == '\n' )) {
		} else {
			*to = *from;
			to++;
		}
		from++;
	}
	*to = *from;
}

char *download_image( char *image_url ) {
	static char tmp_filename[1025];
	int sock;
	struct sockaddr_in sa;
	struct hostent *hinfo;
	char url[1025];
	char result[65535];
	char tmp[65535];
	char *ptr;
	int nr = 0, tr = 0;
	int ifp;
	char host[75] = "";

	DBG( 11, "download_image( '%s' )\n", image_url );

	if ( strncmp( image_url, "http://", 7 ))
		return "";

	strcpy( tmp_filename, "/tmp/gyach.XXXXXX" );

	strncpy( host, image_url + 7, 74 );
	host[74] = '\0';
	ptr = strchr( host, '/' );
	*ptr = '\0';

	ptr = strchr( image_url + 7, '/' );

	hinfo = gethostbyname( host );

	if ( ! hinfo || ( hinfo->h_addrtype != AF_INET )) {
		/* fixme, show an error box */
		return "";
	}

	memset( &sa, 0, sizeof(sa));
	memmove((void*)&sa.sin_addr.s_addr, hinfo->h_addr, hinfo->h_length );

	sa.sin_family = AF_INET;
	sa.sin_port = htons( 80 );

	if (( sock = socket( AF_INET, SOCK_STREAM, 6 )) == -1 ) {
		/* fixme, error message */
		if ( DBG_LEVEL > 10 ) {
			printf( "error in socket()\n" );
		}
		return "";
	} else {
		if ( connect( sock, (struct sockaddr*)&sa, sizeof(sa)) == -1 ) {
			/* fixme, error message */
			if ( DBG_LEVEL > 10 ) {
				printf( "error in connect()\n" );
			}
			close(sock);
			return "";
		}
	}

	snprintf( url, 1024,
		"GET %s HTTP/1.0\r\n"
		"Accept: text/html\r\n"
		"Accept: text/plain\r\n"
		"User-Agent: gyach/?.?\r\n"
		"\r\n"
		"\r\n", ptr );

	write( sock, url, strlen( url ));

	result[0] = '\0';

	tr = 0;
	nr = read( sock, tmp, 512 );
	while(( nr > 0 ) &&
		( tr < 65535 )) {
		tmp[nr] = '\0';
		memcpy( result + tr, tmp, nr );
		tr += nr;

		process_gtk_events();

		nr = read( sock, tmp, 512 );
	}

	close( sock );

	ptr = strstr( result, "\r\n\r\n" );
	ptr += 4;

	if ( strstr( result, "Not Found" ))
		return( "" );

	/* old way when was using mktemp() instead of mkstemp() 
	ifp = open( tmp_filename, O_CREAT | O_TRUNC| O_WRONLY, 0600 );
	*/

	/* new way using mkstemp() */
	if (( ifp = mkstemp( tmp_filename )) == -1 ) {
		fprintf( stderr, "mkstemp( '%s' ) == -1\n", tmp_filename );
		fflush( stderr );
		return( "" );
	}


	write( ifp, ptr, tr - ( ptr - result ));
	close( ifp );

	DBG( 12, "download_image() == '%s'\n", tmp_filename );

	return tmp_filename;
}

int my_system( char *command ) {
	int pid;
	int status;

	if ( ! command ) {
		return( 1 );
	}

	pid = fork();
	if ( pid == -1 ) {
		return( -1 );
	} else if ( pid == 0 ) {
		char *argv[4];
		
		argv[0] = "sh";
		argv[1] = "-c";
		argv[2] = command;
		argv[3] = 0;

		execv( "/bin/sh", argv );
		exit( 127 );
	} else {
		while( 1 ) {
			if ( waitpid( pid, &status, 0 ) == -1 ) {
				return( -1 );
			} else {
				return( status );
			}
		}
	}
}

void *display_url( void *arg ) {
	char *url = (char *)arg;
	char cmd[1025];

	DBG( 11, "display_url( '%s' )\n", (char *)arg );

	if ( strlen( url ) == 0 )
		return( NULL );

	/* quick sanity check */
	if (( strchr( url, ';' )) ||
		( strchr( url, '`' ))) {
		return( NULL );
	}

	snprintf( cmd, 1024, browser_command, url );

	my_system( cmd );

	return( NULL );
}

void find_field( char *str, char *search_str, char *result ) {
	char *ptr;
	char *end;
	char *colon;

	DBG( 21, "find_field( '%s', '%s', %p )\n", str, search_str, result );

	ptr = strstr( str, search_str );
	if ( ptr ) {
		ptr += strlen( search_str );
		while(( *ptr != '\n' ) && ( ptr < ( str + strlen( str ))))
			ptr++;
		while(( *ptr == '\n' ) && ( ptr < ( str + strlen( str ))))
			ptr++;
		end = strchr( ptr, '\n' );
		colon = strchr( ptr, ':' );
		if (( colon ) &&
			( colon < end )) {
			strcpy( result, "" );
		} else {
			strncpy( result, ptr, end - ptr );
			result[end - ptr] = '\0';
		}
	}
}

void display_profile() {
	GdkPixmap	*st_pixmap;
	GdkBitmap	*st_mask;
	GtkStyle	*user_style = NULL;
	GtkWidget	*tmp_widget;

	DBG( 11, "display_profile()\n" );

	if ( ! prof_text ) {
		prof_text = NULL;
		prof_pic = NULL;
		return;
	}

	if ( ! profile_window )
		profile_window = create_profile_window();

	if ( ! profile_style ) {
		profile_style = gtk_style_new();
		profile_style->font = gdk_font_load( "fixed" );
	}

	tmp_widget = lookup_widget( profile_window, "profile_info" );

	gtk_widget_set_style( tmp_widget, profile_style );
	gtk_label_set_text( GTK_LABEL(tmp_widget), prof_text );
	gtk_misc_set_alignment( GTK_MISC(tmp_widget), 0.0f, 0.0f );
	prof_text = NULL;

	if ( prof_pic ) {
		gtk_widget_realize( profile_window );

		tmp_widget = lookup_widget( profile_window, "user_picture" );

		user_style = gtk_widget_get_style( profile_window );

		if ( user_style ) {
			st_pixmap = gdk_pixmap_create_from_xpm( profile_window->window,
				&st_mask, &user_style->bg[GTK_STATE_NORMAL], prof_pic );

			gtk_pixmap_set( GTK_PIXMAP(tmp_widget), st_pixmap, st_mask );
		}

		unlink( prof_pic );
		prof_pic = NULL;
	}

	gtk_widget_show( profile_window );
}

/* this function gets run as a separate thread so it does not hold up chat */
/* while downloading html and image and parsing/displaying */
void *fetch_profile( void *arg ) {
	char *user = strdup( (char *)arg );
	int sock;
	struct sockaddr_in sa;
	struct hostent *hinfo;
	char url[1025];
	char result[32769];
	char tmp[32769];
	char buf[8192];
	char cmd[512];
	int nr = 0, tr = 0;

	char *ptr;
	char *ptr2;
	char pic_url[129] = "";
	char last_updated[21] = "";
	char email[129] = "";
	char realname[51] = "";
	char location[51] = "";
	char age[5] = "";
	char married[21] = "";
	char sex[11] = "";
	char occupation[51] = "";
	char online_now[11] = "";
	char homepage[129] = "";
	char coollink1[129] = "";
	char coollink2[129] = "";
	char coollink3[129] = "";

	DBG( 11, "fetch_profile( '%s' )\n", user );

	if ( ! profile_viewer ) {
		snprintf( homepage, 128, "http://profiles.yahoo.com/%s", user );
		display_url( (void *)homepage );
		free( user );
		return( NULL );
	}

	hinfo = gethostbyname( "profiles.yahoo.com" );

	if ( ! hinfo || ( hinfo->h_addrtype != AF_INET )) {
		/* fixme, show an error box */
		free( user );
		return( NULL );
	}

	memset( &sa, 0, sizeof(sa));
	memmove((void*)&sa.sin_addr.s_addr, hinfo->h_addr, hinfo->h_length );

	sa.sin_family = AF_INET;
	sa.sin_port = htons( 80 );

	if (( sock = socket( AF_INET, SOCK_STREAM, 6 )) == -1 ) {
		/* fixme, error message */
		if ( DBG_LEVEL > 10 ) {
			printf( "error in socket()\n" );
		}
		free( user );
		return( NULL );
	} else {
		if ( connect( sock, (struct sockaddr*)&sa, sizeof(sa)) == -1 ) {
			/* fixme, error message */
			if ( DBG_LEVEL > 10 ) {
				printf( "error in connect()\n" );
			}
			close(sock);
			free( user );
			return( NULL ); 
		}
	}

	snprintf( url, 1024,
		"GET /%s HTTP/1.0\r\n"
		"Accept: text/html\r\n"
		"Accept: text/plain\r\n"
		"User-Agent: gyach/?.?\r\n"
		"\r\n"
		"\r\n", user );

	write( sock, url, strlen( url ));

	result[0] = '\0';

	tr = 0;
	nr = read( sock, tmp, 32768 );
	while( nr > 0 ) {
		tmp[nr] = '\0';
		strcat( result, tmp );
		tr += nr;
		nr = read( sock, tmp, 32768 - tr );
	}

	/* find the url to their pic */
	sprintf( tmp, "alt=\"%s\"", user );
	ptr = tmp;
	while( *ptr )
		*(ptr++) = tolower( *ptr );
	ptr = strstr( result, tmp );
	if ( ptr ) {
		while((( *(ptr-4) != 's' ) || ( *(ptr-3) != 'r' ) ||
			   ( *(ptr-2) != 'c' ) || ( *(ptr-1) != '=' )) &&
			  ( ptr > (result+5))) {
			ptr--;
		}
		if ( *ptr == '"' )
			ptr++;
		/* now we are at the beginning of the url, so strncpy some */
		strncpy( pic_url, ptr, 128 );
		ptr = strchr( pic_url, '"' );
		ptr2 = strchr( pic_url, ' ' );
		if (( ptr && ptr2 ) &&
			( ptr2 < ptr ))
			ptr = ptr2;

		if ( ptr )
			*ptr = '\0';
		else
			pic_url[128] = '\0';
	}

	/* strip html to make the rest easier */
	strip_html_tags( result );

	/* attempt to find some strings */
	find_field( result, "Last Updated:", last_updated );
	find_field( result, "My Email", email );
	find_field( result, "Real Name:", realname );
	find_field( result, "Location:", location );
	find_field( result, "Age:", age );
	find_field( result, "Marital Status:", married );
	find_field( result, "Sex:", sex );
	find_field( result, "Occupation:", occupation );
	find_field( result, "Home Page:", homepage );

	ptr = strstr( result, "I'm online now!" );
	if ( ptr ) {
		strcpy( online_now, "Online now!" );
	} else {
		strcpy( online_now, "OFFline" );
	}

	if ( DBG_LEVEL >= 10 ) {
		printf( "profile info for %s:\n", user );
		printf( "Picture         : '%s'\n", pic_url );
		printf( "Last Updated    : '%s'\n", last_updated );
		printf( "Email           : '%s'\n", email );
		printf( "Real Name       : '%s'\n", realname );
		printf( "Location        : '%s'\n", location );
		printf( "Age             : '%s'\n", age );
		printf( "Married?        : '%s'\n", married );
		printf( "Sex             : '%s'\n", sex );
		printf( "Occupation      : '%s'\n", occupation );
		printf( "Online Now      : '%s'\n", online_now );
		printf( "Home Page       : '%s'\n", homepage );
		printf( "Cool Link #1    : '%s'\n", coollink1 );
		printf( "Cool Link #2    : '%s'\n", coollink2 );
		printf( "Cool Link #3    : '%s'\n", coollink3 );
	}

	close(sock);

	sprintf( buf,
		" Yahoo! ID      : %s\n"
		" Real Name      : %s\n"
		" Last Updated   : %s\n"
		" Location       : %s\n"
		" Age            : %s\n"
		" Marital Status : %s\n"
		" Sex            : %s\n"
		" Occupation     : %s\n"
		" Email          : %s\n"
		" HomePage       : %s\n"
		" Cool Links     : %s\n"
		"                  %s\n"
		"                  %s\n",
		user, realname, last_updated, location, age, married, sex, occupation,
		email, homepage, coollink1, coollink2, coollink3 );

	if ( pic_url[0] ) {
		ptr = download_image( pic_url );

		sprintf( cmd, "convert -geometry 200x200 -colors 256 %s "
			"xpm:%s.xpm", ptr, ptr );
	
		if ( ! my_system( cmd )) {
			sprintf( cmd, "%s.xpm", ptr );
			prof_pic = strdup( cmd );
		}

		unlink( ptr );
	}

	prof_text = strdup( buf );

	free( user );
	return( NULL );
}

void lower_str( char *str ) {
	char *ptr = str;

	DBG( 21, "lower_str( '%s' )\n", str );

	while( *ptr ) {
		*(ptr++) = tolower( *ptr );
	}

	DBG( 22, "RETURN lower_str() == '%s'\n", str );
}

void edit_emote() {
	gchar *listitem[3];
	gchar emotion[128] = "";
	gchar em_everyone[128] = "";
	gchar em_user[128] = "";
	GtkWidget *emote_name;
	GtkWidget *emote_everyone;
	GtkWidget *emote_user;

	listitem[0] = emotion;
	listitem[1] = em_everyone;
	listitem[2] = em_user;

	DBG( 11, "edit_emote()\n" );

	if ( emote_selected > -1 ) {
		gtk_clist_get_text( GTK_CLIST(chat_emotions), emote_selected, 0,
			listitem );
		gtk_clist_get_text( GTK_CLIST(chat_emotions), emote_selected, 1,
			listitem+1 );
		gtk_clist_get_text( GTK_CLIST(chat_emotions), emote_selected, 2,
			listitem+2 );
	}

    if ( ! emote_window ) {
		emote_window = create_emote_window();
		gtk_widget_show( emote_window );
	}

	emote_name = lookup_widget( emote_window, "emote_name" );
	emote_everyone = lookup_widget( emote_window, "emote_everyone" );
	emote_user = lookup_widget( emote_window, "emote_user" );

	gtk_entry_set_text( GTK_ENTRY(emote_name), listitem[0] );
	gtk_entry_set_text( GTK_ENTRY(emote_everyone), listitem[1] );
	gtk_entry_set_text( GTK_ENTRY(emote_user), listitem[2] );
}

int find_user_row( char *user ) {
	char tmp_user[513];
	char tmp_list[513];
	char *listitem[1];
	int i;
	int r;

	DBG( 21, "find_user_row( '%s' )\n", user );

	if ( strlen( user ) > sizeof( tmp_user ))
		return( -1 );

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	r = -1;
	for( i = 0 ; i < GTK_CLIST(chat_users)->rows && r == -1; i++ ) {
		gtk_clist_get_text( GTK_CLIST(chat_users), i, 1, listitem );
		strcpy( tmp_list, listitem[0] );
		lower_str( tmp_list );
		if ( ! strcmp( tmp_list, tmp_user ))
			r = i;
	}

	DBG( 22, "RETURN find_user_row() == %d\n", r );

	return( r );
}

int chars_match( char *str1, char *str2 ) {
	int i = 0;

	DBG( 21, "chars_match( '%s', '%s' )\n", str1, str2 );

	/* if the 2nd string is longer then it's can't be the same word */
	if ( strlen( str2 ) > strlen( str1 ) )
		return( 0 );

	while(( *str1 ) && ( *str2 ) && ( *str1 == *str2 )) {
		i++;
		str1++;
		str2++;
	}

	DBG( 22, "RETURN chars_match() == %d\n", i );

	return( i );
}

/*
 * fixme, this function needs cleaning up probably
 */
char *find_user_best_match( char *part ) {
	static char best_match[129] = "";
	static char last_part[129] = "";
	static int  last_match = 0;
	char tmp_user[129];
	char tmp_list[129];
	char *listitem[1];
	int bm = last_match;

	DBG( 11, "find_user_best_match( '%s' )\n", part );

	if ( ! part[0] ) {
		best_match[0] = '\0';
		last_part[0] = '\0';
		return( best_match );
	}

	if (( ! last_part[0] ) || strncmp( part, last_part, strlen( last_part ))) {
		best_match[0] = '\0';
		last_match = 0;
		strcpy( last_part, part );
	} else {
		strcpy( part, last_part );
		last_match++;
	}

	if ( last_match >= GTK_CLIST(chat_users)->rows )
		last_match = 0;

	strcpy( tmp_user, part );
	lower_str( tmp_user );

	gtk_clist_get_text( GTK_CLIST(chat_users), last_match, 1, listitem );
	strcpy( tmp_list, listitem[0] );
	lower_str( tmp_list );

	while(( strncmp( tmp_list, part, strlen( part ))) &&
		  ( last_match < GTK_CLIST(chat_users)->rows )) {
		last_match++;
		gtk_clist_get_text( GTK_CLIST(chat_users), last_match, 1, listitem );
		strcpy( tmp_list, listitem[0] );
		lower_str( tmp_list );
	}

	/* did we hit the end of the list after starting at the beginning */
	/* if so, return */
	if (( last_match >= GTK_CLIST(chat_users)->rows ) &&
		( bm == 0 )) {
		strcpy( best_match, part );
		last_match = 0;
		return( best_match );
	} 

	/* if we don't match, try to start from the beginning again */
	if ( strncmp( tmp_list, part, strlen( part ))) {
		last_match = 0;
		gtk_clist_get_text( GTK_CLIST(chat_users), last_match, 1, listitem );
		strcpy( tmp_list, listitem[0] );
		lower_str( tmp_list );

		while(( strncmp( tmp_list, part, strlen( part ))) &&
			  ( last_match < bm )) {
			last_match++;
			gtk_clist_get_text( GTK_CLIST(chat_users), last_match, 1,
				listitem );
			strcpy( tmp_list, listitem[0] );
			lower_str( tmp_list );
		}
	}

	if ( strncmp( tmp_list, part, strlen( part ))) {
		strcpy( best_match, part );
		last_match = 0;
	} else {
		strcpy( best_match, tmp_list );
	}
		
	return( best_match );
}

void ignore_add( char *user ) {
	char tmp_user[129];

	if ( ! user ) {
		DBG( 11, "ignore_add( NULL )\n" );
		return;
	} else {
		DBG( 11, "ignore_add( '%s' )\n", user );
	}

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	if ( g_list_find_custom( ignore_list, tmp_user, (GCompareFunc)gstrcmp ))
		return;

	ignore_list = g_list_prepend( ignore_list, strdup( tmp_user ));

	/*
	g_list_sort( ignore_list, (GCompareFunc)gstrcmp );
	*/
}

void ignore_remove( char *user ) {
	GList *this_user;
	char tmp_user[129];
	char *ptr;

	if ( ! user ) {
		DBG( 11, "ignore_remove( NULL )\n" );
		return;
	} else {
		DBG( 11, "ignore_remove( '%s' )\n", user );
	}

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	this_user = g_list_find_custom( ignore_list, tmp_user,
		(GCompareFunc)gstrcmp );

	if ( this_user ) {
		ptr = this_user->data;
		ignore_list = g_list_remove( ignore_list, ptr );
		free( ptr );
	}
}

int ignore_check( char *user ) {
	GList *this_user;
	char tmp_user[129];
	char *ptr;

	if ( ! user ) {
		DBG( 11, "ignore_check( NULL )\n" );
		return( 0 );
	} else {
		DBG( 11, "ignore_check( '%s' )\n", user );
	}

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	/* check for .guest user */
	if (( ignore_guests ) &&
		( ! strcmp( &tmp_user[strlen(tmp_user)-6], ".guest" ))) {
		return( 1 );
	}

	this_user = g_list_find_custom( ignore_list, tmp_user,
		(GCompareFunc)gstrcmp );

	if ( this_user ) {
		/* if we found the user then move them to the front of the list */
		/* for speed in looking them up next time */
		DBG( 13, "ignore_check(), found user\n" );
		ptr = this_user->data;
		DBG( 13, "ignore_check(), removing from list\n" );
		ignore_list = g_list_remove_link( ignore_list, this_user );
		DBG( 13, "ignore_check(), prepending to list\n" );
		ignore_list = g_list_prepend( ignore_list, ptr );
		DBG( 13, "ignore_check(), freeing old list item\n" );
		g_list_free( this_user );

		DBG( 12, "RETURN ignore_check( '%s' ) == 1\n", user );
		return( 1 );
	} else {
		DBG( 12, "RETURN ignore_check( '%s' ) == 0\n", user );
		return( 0 );
	}
}

void regex_add( char *regex ) {
	char tmp_regex[129];

	if ( ! regex ) {
		DBG( 11, "regex_add( NULL )\n" );
		return;
	} else {
		DBG( 11, "regex_add( '%s' )\n", regex );
	}

	strcpy( tmp_regex, regex );
	lower_str( tmp_regex );

	if ( g_list_find_custom( regex_list, tmp_regex, (GCompareFunc)gstrcmp ))
		return;

	regex_list = g_list_prepend( regex_list, strdup( tmp_regex ));
	/*
	g_list_sort( regex_list, (GCompareFunc)gstrcmp );
	*/
}

void regex_remove( char *regex ) {
	GList *this_regex;
	char tmp_regex[129];
	char *ptr;

	if ( ! regex ) {
		DBG( 11, "regex_remove( NULL )\n" );
		return;
	} else {
		DBG( 11, "regex_remove( '%s' )\n", regex );
	}

	strcpy( tmp_regex, regex );
	lower_str( tmp_regex );

	this_regex = g_list_find_custom( regex_list, tmp_regex,
		(GCompareFunc)gstrcmp );

	if ( this_regex ) {
		ptr = this_regex->data;
		regex_list = g_list_remove( regex_list, ptr );
		free( ptr );
	}
}

int regex_check( char *regex ) {
	GList *this_regex;
	char tmp_regex[129];

	if ( ! regex ) {
		DBG( 11, "regex_check( NULL )\n" );
		return( 0 );
	} else {
		DBG( 11, "regex_check( '%s' )\n", regex );
	}

	strcpy( tmp_regex, regex );
	lower_str( tmp_regex );

	this_regex = g_list_find_custom( regex_list, tmp_regex,
		(GCompareFunc)gstrcmp );

	if ( this_regex )
		return( 1 );
	else
		return( 0 );
}

int auto_ignore_check( char *str ) {
	GList *this_regex;
	regex_t ai_regex;
	

	if ( ! str ) {
		DBG( 11, "auto_ignore_check( NULL )\n" );
		return( 0 );
	} else {
		DBG( 11, "auto_ignore_check( '%s' )\n", str );
	}

	this_regex = regex_list;
	while( this_regex ) {
		regcomp( &ai_regex, this_regex->data, REG_ICASE | REG_NOSUB );
		if ( ! regexec( &ai_regex, str, 0, NULL, 0 )) {
			regfree( &ai_regex );
			return( 1 );
		}
		regfree( &ai_regex );

		this_regex = g_list_next( this_regex );
	}

	return( 0 );
}

void ignore_load() {
	GList *this_link;
	char if_filename[1024];
	char buf[129];
	FILE *fp;

	DBG( 11, "ignore_load()\n" );

	if ( cfg_dir[ 0 ] == '\0' )
		sprintf( cfg_dir, "%s/.gyach", getenv( "HOME" ));

	sprintf( if_filename, "%s/ignore.list", cfg_dir );

	fp = fopen( if_filename, "r" );

	if ( fp ) {
		/* clear the list if there already is one */
		this_link = ignore_list;
		while( this_link ) {
			free( this_link->data );
			this_link = g_list_next( this_link );
		}

		g_list_free( ignore_list );

		ignore_list = NULL;

		/* load the new list */
		while( fgets( buf, 128, fp )) {
			/* chop the newline */
			buf[strlen(buf)-1] = '\0';

			ignore_list = g_list_append( ignore_list, strdup( buf ));
		}

		fclose( fp );

		/*
		g_list_sort( ignore_list, (GCompareFunc)gstrcmp );
		*/
	}

	sprintf( if_filename, "%s/ignore.regex", cfg_dir );

	fp = fopen( if_filename, "r" );

	if ( fp ) {
		/* clear the list if there already is one */
		this_link = regex_list;
		while( this_link ) {
			free( this_link->data );
			this_link = g_list_next( this_link );
		}

		g_list_free( regex_list );

		regex_list = NULL;

		/* load the new list */
		while( fgets( buf, 128, fp )) {
			/* chop the newline */
			buf[strlen(buf)-1] = '\0';

			regex_list = g_list_append( regex_list, strdup( buf ));
		}

		fclose( fp );

		/*
		g_list_sort( regex_list, (GCompareFunc)gstrcmp );
		*/
	}
}

void ignore_save() {
	GList *this_user;
	GList *this_regex;
	char if_filename[1024];
	FILE *fp;

	DBG( 11, "ignore_save()\n" );

	if ( cfg_dir[ 0 ] == '\0' )
		sprintf( cfg_dir, "%s/.gyach", getenv( "HOME" ));

	sprintf( if_filename, "%s/ignore.list", cfg_dir );

	fp = fopen( if_filename, "w" );

	if ( fp ) {
		this_user = ignore_list;
		while( this_user ) {
			fprintf( fp, "%s\n", (char *)this_user->data );

			this_user = g_list_next( this_user );
		}

		fclose( fp );
	}

	sprintf( if_filename, "%s/ignore.regex", cfg_dir );

	fp = fopen( if_filename, "w" );

	if ( fp ) {
		this_regex = regex_list;
		while( this_regex ) {
			fprintf( fp, "%s\n", (char *)this_regex->data );
			this_regex = g_list_next( this_regex );
		}

		fclose( fp );
	}
}

void ignore_toggle( char *user ) {
	GList *this_user;
	int r;
	int i;
	char buf[129] = "";
	GtkWidget *tmp_widget;
	GtkStyle *user_style;
#ifdef SHOW_STATUS_PIXMAPS
	GdkPixmap *st_pixmap = NULL;
	GdkBitmap *st_mask = NULL;
#else
	char status[5] = "";
#endif

	DBG( 11, "ignore_toggle( '%s' )\n", user );

	r = find_user_row( user );

	tmp_widget = lookup_widget( chat_window, "users" );
	user_style = gtk_widget_get_style( tmp_widget );

	if ( user[0] ) {
		if ( ignore_check( user )) {
			ignore_remove( user );
			if ( r >= 0 ) {
#ifdef SHOW_STATUS_PIXMAPS
				st_pixmap = gdk_pixmap_create_from_xpm_d( tmp_widget->window,
					&st_mask, &user_style->bg[GTK_STATE_NORMAL],
					(gchar **)pixmap_status_here );
#else
				strcpy( status, "" );
#endif
			}
			sprintf( buf, "*** no longer ignoring %s%s%s ***\n",
				YAHOO_COLOR_BLUE, user, YAHOO_COLOR_BLACK );
		} else {
			ignore_add( user );
			if ( r >= 0 ) {
#ifdef SHOW_STATUS_PIXMAPS
				st_pixmap = gdk_pixmap_create_from_xpm_d( tmp_widget->window,
					&st_mask, &user_style->bg[GTK_STATE_NORMAL],
					(gchar **)pixmap_status_ignore );
#else
				strcpy( status, "IG" );
#endif
			}
			sprintf( buf, "*** ignoring %s%s%s ***\n",
				YAHOO_COLOR_BLUE, user, YAHOO_COLOR_BLACK );
		}
		append_to_textbox( chat_window, chat_text, buf );
	} else {
		append_to_textbox( chat_window, chat_text,
			"*** Ignoring the following users:\n" );
		i = 0;
		this_user = ignore_list;
		while( this_user ) {
			sprintf( buf, "\t%s\n", (char *)this_user->data );
			append_to_textbox( chat_window, chat_text, buf );
			i++;
			this_user = g_list_next( this_user );
		}
		if ( i ) {
			sprintf( buf, "*** Ignoring %d users ***\n", i );
			append_to_textbox( chat_window, chat_text, buf );
		} else {
			append_to_textbox( chat_window, chat_text, "\tNo Users Ignored\n" );
		}
	}

	ignore_save();

	if ( r < 0 )
		return;

#ifdef SHOW_STATUS_PIXMAPS
	gtk_clist_set_pixmap(GTK_CLIST(chat_users), r, 0, st_pixmap, st_mask );
	gdk_pixmap_unref( st_pixmap );
	gdk_bitmap_unref( st_mask );
#else
	gtk_clist_set_text(GTK_CLIST(chat_users), r, 0, status );
#endif
}

void mute_add( char *user ) {
	char tmp_user[129];

	if ( ! user ) {
		DBG( 11, "mute_add( NULL )\n" );
		return;
	} else {
		DBG( 11, "mute_add( '%s' )\n", user );
	}

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	if ( g_list_find_custom( mute_list, tmp_user, (GCompareFunc)gstrcmp ))
		return;

	mute_list = g_list_prepend( mute_list, strdup( tmp_user ));
	/*
	g_list_sort( mute_list, (GCompareFunc)gstrcmp );
	*/
}

void mute_remove( char *user ) {
	GList *this_user;
	char tmp_user[129];
	char *ptr;

	if ( ! user ) {
		DBG( 11, "mute_remove( NULL )\n" );
		return;
	} else {
		DBG( 11, "mute_remove( '%s' )\n", user );
	}

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	this_user = g_list_find_custom( mute_list, tmp_user,
		(GCompareFunc)gstrcmp );

	if ( this_user ) {
		ptr = this_user->data;
		mute_list = g_list_remove( mute_list, ptr );
		free( ptr );
	}
}

int mute_check( char *user ) {
	GList *this_user;
	char tmp_user[129];
	char *ptr;

	if ( ! user ) {
		DBG( 11, "mute_check( NULL )\n" );
		return( 0 );
	} else {
		DBG( 11, "mute_check( '%s' )\n", user );
	}

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	this_user = g_list_find_custom( mute_list, tmp_user,
		(GCompareFunc)gstrcmp );

	if ( this_user ) {
		/* if we found the user then move them to the front of the list */
		/* for speed in looking them up next time */
		ptr = this_user->data;
		mute_list = g_list_remove_link( mute_list, this_user );
		mute_list = g_list_prepend( mute_list, ptr );
		g_list_free( this_user );

		return( 1 );
	} else {
		return( 0 );
	}
}

void mute_toggle( char *user ) {
	GList *this_user;
	int r;
	int i;
	char buf[129] = "";
	GtkWidget *tmp_widget;
	GtkStyle *user_style;
#ifdef SHOW_STATUS_PIXMAPS
	GdkPixmap *st_pixmap = NULL;
	GdkBitmap *st_mask = NULL;
#else
	char status[5] = "";
#endif

	DBG( 11, "mute_toggle( '%s' )\n", user );

	r = find_user_row( user );

	tmp_widget = lookup_widget( chat_window, "users" );
	user_style = gtk_widget_get_style( tmp_widget );

	if ( user[0] ) {
		if ( mute_check( user )) {
			mute_remove( user );
			if ( r >= 0 ) {
#ifdef SHOW_STATUS_PIXMAPS
				st_pixmap = gdk_pixmap_create_from_xpm_d( tmp_widget->window,
					&st_mask, &user_style->bg[GTK_STATE_NORMAL],
					(gchar **)pixmap_status_here );
#else
				strcpy( status, "" );
#endif
			}
			sprintf( buf, "*** no longer muting %s%s%s ***\n",
				YAHOO_COLOR_BLUE, user, YAHOO_COLOR_BLACK );
		} else {
			mute_add( user );
			if ( r >= 0 ) {
#ifdef SHOW_STATUS_PIXMAPS
				st_pixmap = gdk_pixmap_create_from_xpm_d( tmp_widget->window,
					&st_mask, &user_style->bg[GTK_STATE_NORMAL],
					(gchar **)pixmap_status_mute );
#else
				strcpy( status, "MU" );
#endif
			}
			sprintf( buf, "*** muting %s%s%s ***\n",
				YAHOO_COLOR_BLUE, user, YAHOO_COLOR_BLACK );
		}
		append_to_textbox( chat_window, chat_text, buf );
	} else {
		append_to_textbox( chat_window, chat_text,
			"*** Muting the following users:\n" );
		i = 0;
		this_user = mute_list;
		while( this_user ) {
			sprintf( buf, "\t%s\n", (char *)this_user->data );
			append_to_textbox( chat_window, chat_text, buf );
			i++;
			this_user = g_list_next( this_user );
		}
		if ( i ) {
			sprintf( buf, "*** Muting %d users ***\n", i );
			append_to_textbox( chat_window, chat_text, buf );
		} else {
			append_to_textbox( chat_window, chat_text, "\tNo Users Muted\n" );
		}
	}

	if ( r < 0 )
		return;

#ifdef SHOW_STATUS_PIXMAPS
	gtk_clist_set_pixmap(GTK_CLIST(chat_users), r, 0, st_pixmap, st_mask );
	gdk_pixmap_unref( st_pixmap );
	gdk_bitmap_unref( st_mask );
#else
	gtk_clist_set_text(GTK_CLIST(chat_users), r, 0, status );
#endif
}

void regex_toggle( char *regex ) {
	char buf[129] = "";

	DBG( 11, "regex_toggle( '%s' )\n", regex );

	if ( regex[0] ) {
		if ( regex_check( regex )) {
			regex_remove( regex );
			sprintf( buf, "*** no longer auto-ignoring on regex '%s%s%s' ***\n",
				YAHOO_COLOR_BLUE, regex, YAHOO_COLOR_BLACK );
		} else {
			regex_add( regex );
			sprintf( buf, "*** auto-ignoring on regex '%s%s%s' ***\n",
				YAHOO_COLOR_BLUE, regex, YAHOO_COLOR_BLACK );
		}
		append_to_textbox( chat_window, chat_text, buf );
	}

	ignore_save();
}

/*
 * dump a yin packet out to stdout for testing purposes
 */
void dump_packet() {
	int i = 0;
	int x = 0;
	int y = 0;
	char *ptr;

	DBG( 21, "dump_packet()\n" );

	fprintf( stderr, "type = 0x%x, flag = %d, size = %d\n",
		GlobalSessionHook->yin.type,
		GlobalSessionHook->yin.flag,
		GlobalSessionHook->yin.size );

	ptr = GlobalSessionHook->yin.data;
	while( i < GlobalSessionHook->yin.size ) {
		if ( x < 16 ) {
			fprintf( stderr, "%02x ", *(ptr+i) & 0xFF );
			x++;
			i++;
		} else {
			fprintf( stderr, "   " );
			for( ; x > 0 ; x-- ) {
				if ( isgraph( *(ptr+i-x) ) || *(ptr+i-x)  == ' ' ) {
					fprintf( stderr, "%c", *(ptr+i-x) );
				} else {
					fprintf( stderr, "." );
				}
			}

			fprintf( stderr, "\n" );
			x = 0;
		}
	}
	for( y = x; y < 16 ; y++ ) {
		fprintf( stderr, "   " );
	}
	fprintf( stderr, "   " );
	for( ; x > 0 ; x-- ) {
		if ( isgraph( *(ptr+i-x) ) || *(ptr+i-x)  == ' ' ) {
			fprintf( stderr, "%c", *(ptr+i-x) );
		} else {
			fprintf( stderr, "." );
		}
	}

	fprintf( stderr, "\n" );
}

/*
 * attempt to highlight urls if they aren't already highlighted along with
 * turning html style tags into esc sequences.
 */
void convert_tags( char *str ) {
	char tmp[4097];
	char *from;
	char *to;
	char *ptr;
	int len = 0;

	DBG( 21, "convert_tags( '%s' )\n", str );

	/* return if too long */
	if ( strlen( str ) > 4096 )
		return;

	/* convert html tags to ESC sequences */
	if ( strchr( str, '<' )) {
		memset( tmp, 0, sizeof( tmp ));

		from = str;
		to = tmp;

		ptr = strchr( from, '<' );
		while( ptr ) {
			len = ptr - from;
			strncpy( to, from, len );
			to += len;
			*to = '\0';

			from += len;

			if ( !strncasecmp( from, "<black>", 7 )) {
				if ( show_html ) {
					strcat( to, YAHOO_COLOR_BLACK );
					to += strlen( YAHOO_COLOR_BLACK );
				}
				from += 7;
			} else if ( !strncasecmp( from, "<blue>", 6 )) {
				if ( show_html ) {
					strcat( to, YAHOO_COLOR_BLUE );
					to += strlen( YAHOO_COLOR_BLUE );
				}
				from += 6;
			} else if ( !strncasecmp( from, "<cyan>", 6 )) {
				if ( show_html ) {
					strcat( to, YAHOO_COLOR_CYAN );
					to += strlen( YAHOO_COLOR_CYAN );
				}
				from += 6;
			} else if ( !strncasecmp( from, "<gray>", 6 )) {
				if ( show_html ) {
					strcat( to, YAHOO_COLOR_GRAY );
					to += strlen( YAHOO_COLOR_GRAY );
				}
				from += 6;
			} else if ( !strncasecmp( from, "<green>", 7 )) {
				if ( show_html ) {
					strcat( to, YAHOO_COLOR_GREEN );
					to += strlen( YAHOO_COLOR_GREEN );
				}
				from += 7;
			} else if ( !strncasecmp( from, "<pink>", 6 )) {
				if ( show_html ) {
					strcat( to, YAHOO_COLOR_PINK );
					to += strlen( YAHOO_COLOR_PINK );
				}
				from += 6;
			} else if ( !strncasecmp( from, "<purple>", 8 )) {
				if ( show_html ) {
					strcat( to, YAHOO_COLOR_PURPLE );
					to += strlen( YAHOO_COLOR_PURPLE );
				}
				from += 8;
			} else if ( !strncasecmp( from, "<orange>", 8 )) {
				if ( show_html ) {
					strcat( to, YAHOO_COLOR_ORANGE );
					to += strlen( YAHOO_COLOR_ORANGE );
				}
				from += 8;
			} else if ( !strncasecmp( from, "<red>", 5 )) {
				if ( show_html ) {
					strcat( to, YAHOO_COLOR_RED );
					to += strlen( YAHOO_COLOR_RED );
				}
				from += 5;
			} else if ( !strncasecmp( from, "<yellow>", 8 )) {
				if ( show_html ) {
					strcat( to, "\033[#ffff00m" );
					to += strlen( "\033[#ffff00m" );
				}
				from += 5;
			} else if ( !strncasecmp( from, "<i>", 3 )) {
			} else if ( !strncasecmp( from, "<i>", 3 )) {
				if ( show_html ) {
					strcat( to, YAHOO_STYLE_ITALICON );
					to += strlen( YAHOO_STYLE_ITALICON );
				}
				from += 3;
			} else if ( !strncasecmp( from, "</i>", 4 )) {
				if ( show_html ) {
					strcat( to, YAHOO_STYLE_ITALICOFF );
					to += strlen( YAHOO_STYLE_ITALICOFF );
				}
				from += 4;
			} else if ( !strncasecmp( from, "<b>", 3 )) {
				if ( show_html ) {
					strcat( to, YAHOO_STYLE_BOLDON );
					to += strlen( YAHOO_STYLE_BOLDON );
				}
				from += 3;
			} else if ( !strncasecmp( from, "</b>", 4 )) {
				if ( show_html ) {
					strcat( to, YAHOO_STYLE_BOLDOFF );
					to += strlen( YAHOO_STYLE_BOLDOFF );
				}
				from += 4;
			} else if ( !strncasecmp( from, "</font>", 7 )) {
				/* skip ending font tags */
				from += 7;
			} else if ( !strncasecmp( from, "<alt", 4 )) {
				/* skip alt tags */
				from += 4;
				while(( *from ) && ( *from != '>' ))
					from++;
				from++;
			} else if ( !strncasecmp( from, "</alt>", 6 )) {
				/* skip ending alt tags */
				from += 6;
			} else if ( !strncasecmp( from, "<url=", 5 )) {
				/* skip beginning url tags */
				ptr = from + 5;
				while(( *ptr ) && ( *ptr != '>' )) {
					*(to++) = *(ptr)++;
				}
				if ( *ptr )
					from = ptr + 1;
				*(to++) = ' ';
				*(to++) = '(';
			} else if ( !strncasecmp( from, "</url>", 6 )) {
				/* skip ending url tags */
				*(to++) = ')';
				from += 6;
			} else if ( !strncasecmp( from, "<snd", 4 )) {
				/* skip beginning snd tags */
				ptr = from + 4;
				while(( *ptr ) && ( *ptr != '>' ))
					ptr++;
				if ( *ptr )
					from = ptr + 1;
			} else if ( !strncasecmp( from, "</snd>", 6 )) {
				/* skip ending snd tags */
				from += 6;
			} else if ( !strncasecmp( from, "<fade", 5 )) {
				/* skip beginning fade tags */
				ptr = from + 5;
				while(( *ptr ) && ( *ptr != '>' ))
					ptr++;
				if ( *ptr )
					from = ptr + 1;
			} else if ( !strncasecmp( from, "</fade>", 7 )) {
				/* skip ending fade tags */
				from += 7;
			} else {
				*(to++) = *(from++);
			}

			ptr = strchr( from, '<' );
		}

		strcat( to, from );
		to += strlen( from );
		*to = '\0';
		strcpy( str, tmp );
	}

	/* return if no urls */
	if ( ! strstr( str, "http://" ))
		return;

	/* return if already marked */
	if ( strstr( str, YAHOO_STYLE_URLON ))
		return;

	/* return if not showing html */
	if ( ! show_html )
		return;

	memset( tmp, 0, sizeof( tmp ));

	from = str;
	to = tmp;

	ptr = strstr( str, "http://" );

	while( ptr ) {
		len = ptr - from;
		strncpy( to, from, len );
		to += len;
		*to = '\0';

		from += len;

		strcat( to, YAHOO_STYLE_URLON );
		to += strlen( YAHOO_STYLE_URLON );

		while(( *from ) && ( ! isspace( *from ))) {
			*(to++) = *(from++);
		}
		*to = '\0';

		strcat( to, YAHOO_STYLE_URLOFF );
		to += strlen( YAHOO_STYLE_URLOFF );

		ptr = strstr( from, "http://" );
	}

	strcat( to, from );
	to += strlen( from );
	*to = '\0';
	strcpy( str, tmp );
}

/*
 * strip junk (like "<snd" directives out of text
 */
void strip_junk( char *str ) {
	DBG( 21, "strip_junk( '%s' )\n", str );

	/* fixme */
}

void tab_complete_username() {
	char *text = gtk_entry_get_text( GTK_ENTRY(chat_entry));
	char new[1025];
	char *ptr;

	DBG( 11, "tab_complete_username()\n" );

	if ( ! strlen( text ))
		return;

	strcpy( new, text );
	ptr = new + strlen( new ) - 1;

	while(( ptr > new ) && ( ! isspace( *ptr )))
		ptr--;

	if ( isspace( *ptr ))
		ptr++;

	if ( *ptr == '/' ) {
	} else {
		lower_str( ptr );
		strcpy( ptr, find_user_best_match( ptr ));
	}

	gtk_entry_set_text( GTK_ENTRY(chat_entry), new );
	gtk_entry_set_position( GTK_ENTRY(chat_entry), strlen(new));
}

void capture_text_to_file( char *filename ) {
	FILE *fp;
	int  length;
	int  i;
	gchar ch;
	time_t time_now;

	DBG( 11, "capture_text_to_file( '%s' )\n", filename );

	fp = fopen( filename, "ab" );

	if ( ! fp )
		return;

	length = gtk_text_get_length( GTK_TEXT(chat_text) );

	time_now = time(NULL);

	fprintf( fp, "Captured at: %s", ctime(&time_now) );
	fprintf( fp, "-------------------------------------\n" );

	for( i = 0; i < length; i++ ) {
		ch = GTK_TEXT_INDEX( GTK_TEXT(chat_text), i );
		fwrite( &ch, 1, 1, fp );
	}

	fprintf( fp, "\n" );
	fclose( fp );
}

void set_setup_options() {
	GtkWidget *tmp_widget;
	GList *this_regex;
	GList *this_user;
	GList *this_login;
	GList *this_favroom;
	GList *this_friend;
	gchar *listitem[2];
	int row;

	DBG( 11, "set_setup_options()\n" );

	/* show the correct settings */
	tmp_widget = lookup_widget( setup_window, "auto_login" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), auto_login );

	tmp_widget = lookup_widget( setup_window, "blended_colors" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_blended_colors );

	tmp_widget = lookup_widget( setup_window, "packet_debugging" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), debug_packets );

	if ( auto_away ) {
		tmp_widget = lookup_widget( setup_window, "auto_away" );
		gtk_toggle_button_set_active(
			&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), 1 );

		tmp_widget = lookup_widget( setup_window, "auto_away_time" );
		gtk_spin_button_set_value( GTK_SPIN_BUTTON( tmp_widget ),
			(float)auto_away );
	}

	tmp_widget = lookup_widget( setup_window, "custom_away_message" );
	if ( custom_away_message )
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), custom_away_message );
	else
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), "" );

	tmp_widget = lookup_widget( setup_window, "browser_command" );
	if ( browser_command )
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), browser_command );
	else
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), "" );

	tmp_widget = lookup_widget( setup_window, "filter_command" );
	if ( filter_command )
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), filter_command );
	else
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), "" );
	

	tmp_widget = lookup_widget( setup_window, "pm_brings_back" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), pm_brings_back );

	tmp_widget = lookup_widget( setup_window, "disp_auto_ignored" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), disp_auto_ignored );

	tmp_widget = lookup_widget( setup_window, "remember_position" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), remember_position );

	tmp_widget = lookup_widget( setup_window, "remember_password" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), remember_password );

	tmp_widget = lookup_widget( setup_window, "auto_raise_pm" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), auto_raise_pm );

	tmp_widget = lookup_widget( setup_window, "auto_close_roomlist" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), auto_close_roomlist );

	tmp_widget = lookup_widget( setup_window, "chat_timestamp" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), chat_timestamp );

	tmp_widget = lookup_widget( setup_window, "chat_timestamp_pm" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), chat_timestamp_pm );

	tmp_widget = lookup_widget( setup_window, "auto_reply_when_away" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), auto_reply_when_away );

	tmp_widget = lookup_widget( setup_window, "auto_reply_msg" );
	if ( auto_reply_msg )
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), auto_reply_msg );
	else
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), "" );

	if ( limit_lfs ) {
		tmp_widget = lookup_widget( setup_window, "limit_lfs" );
		gtk_toggle_button_set_active(
			&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), 1 );

		tmp_widget = lookup_widget( setup_window, "limit_lfs_count" );
		gtk_spin_button_set_value( GTK_SPIN_BUTTON( tmp_widget ),
			(float)limit_lfs );
	}

	tmp_widget = lookup_widget( setup_window, "style_bold" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), use_bold );

	tmp_widget = lookup_widget( setup_window, "style_underline" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), use_underline );

	tmp_widget = lookup_widget( setup_window, "style_italic" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), use_italics );

	if ( buffer_size ) {
		tmp_widget = lookup_widget( setup_window, "limit_buffer_size" );
		gtk_toggle_button_set_active(
			&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), 1 );

		tmp_widget = lookup_widget( setup_window, "buffer_size" );
		gtk_spin_button_set_value( GTK_SPIN_BUTTON( tmp_widget ),
			(float)buffer_size );
	}

	if ( ping_minutes ) {
		tmp_widget = lookup_widget( setup_window, "use_ping_time" );
		gtk_toggle_button_set_active(
			&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), 1 );

		tmp_widget = lookup_widget( setup_window, "ping_minutes" );
		gtk_spin_button_set_value( GTK_SPIN_BUTTON( tmp_widget ),
			(float)ping_minutes );
	}

	if ( history_limit ) {
		tmp_widget = lookup_widget( setup_window, "use_history_limit" );
		gtk_toggle_button_set_active(
			&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), 1 );

		tmp_widget = lookup_widget( setup_window, "history_limit" );
		gtk_spin_button_set_value( GTK_SPIN_BUTTON( tmp_widget ),
			(float)history_limit );
	}

	tmp_widget = lookup_widget( setup_window, "force_lowercase" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), force_lowercase );

	tmp_widget = lookup_widget( setup_window, "suppress_mult" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), suppress_mult );

	tmp_widget = lookup_widget( setup_window, "ignore_guests" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), ignore_guests );

	tmp_widget = lookup_widget( setup_window, "auto_reconnect" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), auto_reconnect );

	tmp_widget = lookup_widget( setup_window, "show_colors" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_colors );

	tmp_widget = lookup_widget( setup_window, "show_fonts" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_fonts );

	tmp_widget = lookup_widget( setup_window, "show_html" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_html );

	tmp_widget = lookup_widget( setup_window, "show_statuses" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_statuses );

	tmp_widget = lookup_widget( setup_window, "show_enters" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_enters );

	tmp_widget = lookup_widget( setup_window, "profile_viewer" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), profile_viewer );

	tmp_widget = lookup_widget( setup_window, "ignore_on_mults" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), ignore_on_mults );

	tmp_widget = lookup_widget( setup_window, "pm_in_sep_windows" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), pm_in_sep_windows );

	tmp_widget = lookup_widget( setup_window, "fontselection1" );
	gtk_font_selection_set_font_name(
		GTK_FONT_SELECTION(tmp_widget), font_name );

	/* setup current color */
	if ( ! use_color )
		use_color = strdup( "black" );

	tmp_widget = lookup_widget( setup_window, "color_black" );
	GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active = 0;

	tmp_widget = NULL;
	if ( ! strcmp( use_color, "black" )) {
		tmp_widget = lookup_widget( setup_window, "color_black" );
	} else if ( ! strcmp( use_color, "red" )) {
		tmp_widget = lookup_widget( setup_window, "color_red" );
	} else if ( ! strcmp( use_color, "blue" )) {
		tmp_widget = lookup_widget( setup_window, "color_blue" );
	} else if ( ! strcmp( use_color, "green" )) {
		tmp_widget = lookup_widget( setup_window, "color_green" );
	} else if ( ! strcmp( use_color, "yellow" )) {
		tmp_widget = lookup_widget( setup_window, "color_yellow" );
	} else if ( ! strcmp( use_color, "purple" )) {
		tmp_widget = lookup_widget( setup_window, "color_purple" );
	} else if ( ! strcmp( use_color, "cyan" )) {
		tmp_widget = lookup_widget( setup_window, "color_cyan" );
	} else if ( ! strcmp( use_color, "orange" )) {
		tmp_widget = lookup_widget( setup_window, "color_orange" );
	} else if ( ! strcmp( use_color, "gray" )) {
		tmp_widget = lookup_widget( setup_window, "color_gray" );
/* fixme, support for RGB colors
	} else if ( use_color[0] == '#' ) {
*/
	} else {
		tmp_widget = lookup_widget( setup_window, "color_black" );
	}
	GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active = 1;

	/* loop through and set lists */
	/* Yahoo! login list */
	tmp_widget = lookup_widget( setup_window, "login_list" );
	gtk_clist_freeze(GTK_CLIST(tmp_widget));
	this_login = login_list;
	while( this_login ) {
		listitem[0] = this_login->data;
		listitem[1] = NULL;
		row = gtk_clist_append(GTK_CLIST(tmp_widget), listitem );
		gtk_clist_set_text(GTK_CLIST(tmp_widget), row, 1, "" );

		this_login = g_list_next( this_login );
	}
	gtk_clist_thaw(GTK_CLIST(tmp_widget));

	/* then the favroom list */
	tmp_widget = lookup_widget( setup_window, "favroom_list" );
	gtk_clist_freeze(GTK_CLIST(tmp_widget));
	this_favroom = favroom_list;
	while( this_favroom ) {
		listitem[0] = this_favroom->data;
		listitem[1] = NULL;
		row = gtk_clist_append(GTK_CLIST(tmp_widget), listitem );
		gtk_clist_set_text(GTK_CLIST(tmp_widget), row, 1, "" );

		this_favroom = g_list_next( this_favroom );
	}
	gtk_clist_thaw(GTK_CLIST(tmp_widget));

	/* then the regex list */
	tmp_widget = lookup_widget( setup_window, "regex_list" );
	gtk_clist_set_sort_column(GTK_CLIST(tmp_widget), 0 );
	gtk_clist_set_auto_sort(GTK_CLIST(tmp_widget), 1 );
	gtk_clist_freeze(GTK_CLIST(tmp_widget));

	this_regex = regex_list;
	while( this_regex ) {
		listitem[0] = this_regex->data;
		listitem[1] = NULL;
		row = gtk_clist_insert(GTK_CLIST(tmp_widget), 0, listitem );
		gtk_clist_set_text(GTK_CLIST(tmp_widget), row, 1, "" );

		this_regex = g_list_next( this_regex );
	}
	gtk_clist_thaw(GTK_CLIST(tmp_widget));

	/* then the ignore list */
	tmp_widget = lookup_widget( setup_window, "ignore_list" );
	gtk_clist_set_sort_column(GTK_CLIST(tmp_widget), 0 );
	gtk_clist_set_auto_sort(GTK_CLIST(tmp_widget), 1 );
	gtk_clist_freeze(GTK_CLIST(tmp_widget));
	this_user = ignore_list;
	while( this_user ) {
		listitem[0] = this_user->data;
		row = gtk_clist_insert(GTK_CLIST(tmp_widget), 0, listitem );
		this_user = g_list_next( this_user );
	}
	gtk_clist_thaw(GTK_CLIST(tmp_widget));

	/* now the friend list */
	tmp_widget = lookup_widget( setup_window, "friend_list" );
	gtk_clist_set_sort_column(GTK_CLIST(tmp_widget), 0 );
	gtk_clist_set_auto_sort(GTK_CLIST(tmp_widget), 1 );
	gtk_clist_freeze(GTK_CLIST(tmp_widget));

	this_friend = friend_list;
	while( this_friend ) {
		listitem[0] = this_friend->data;
		row = gtk_clist_insert(GTK_CLIST(tmp_widget), 0, listitem );
		this_friend = g_list_next( this_friend );
	}
	gtk_clist_thaw(GTK_CLIST(tmp_widget));
}

void get_setup_options() {
	GtkWidget *tmp_widget;
	GList *list_ptr;
	char *ptr;
	char tmp_filename[257];
	gchar *tmp_str;
	int i;
	FILE *fp;

	DBG( 11, "get_setup_options()\n" );

	/* gather & save the settings */
	tmp_widget = lookup_widget( setup_window, "auto_login" );
	auto_login = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "blended_colors" );
	show_blended_colors = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "packet_debugging" );
	debug_packets = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "auto_away" );
	auto_away = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	if ( auto_away ) {
		tmp_widget = lookup_widget( setup_window, "auto_away_time" );
		auto_away = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON( tmp_widget ));
	}

	tmp_widget = lookup_widget( setup_window, "custom_away_message" );
	if ( custom_away_message ) {
		free( custom_away_message );
		custom_away_message = NULL;
	}
	custom_away_message = strdup(gtk_entry_get_text(GTK_ENTRY(tmp_widget)));

	tmp_widget = lookup_widget( setup_window, "browser_command" );
	if ( browser_command ) {
		free( browser_command );
		browser_command = NULL;
	}
	browser_command = strdup(gtk_entry_get_text(GTK_ENTRY(tmp_widget)));

	if ( filter_command ) {
		free( filter_command );
		filter_command = NULL;
	}

	tmp_widget = lookup_widget( setup_window, "filter_command" );
	filter_command = strdup(gtk_entry_get_text(GTK_ENTRY(tmp_widget)));
	if ( ! strcmp( filter_command, "" )) {
		free( filter_command );
		filter_command = NULL;
	}
	
	tmp_widget = lookup_widget( setup_window, "pm_brings_back" );
	pm_brings_back = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "disp_auto_ignored" );
	disp_auto_ignored = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "remember_position" );
	remember_position = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "remember_password" );
	remember_password = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "limit_lfs" );
	limit_lfs = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	if ( limit_lfs ) {
		tmp_widget = lookup_widget( setup_window, "limit_lfs_count" );
		limit_lfs = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON( tmp_widget ));
	}

	tmp_widget = lookup_widget( setup_window, "auto_raise_pm" );
	auto_raise_pm = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "auto_close_roomlist" );
	auto_close_roomlist = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "chat_timestamp" );
	chat_timestamp = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "chat_timestamp_pm" );
	chat_timestamp_pm = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "auto_reply_when_away" );
	auto_reply_when_away = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	if ( auto_reply_msg ) {
		free( auto_reply_msg );
		auto_reply_msg = NULL;
	}

	if ( auto_reply_when_away ) {
		tmp_widget = lookup_widget( setup_window, "auto_reply_msg" );
		auto_reply_msg = strdup(gtk_entry_get_text(GTK_ENTRY(tmp_widget)));
	}

	tmp_widget = lookup_widget( setup_window, "style_bold" );
	use_bold = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "style_underline" );
	use_underline = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "style_italic" );
	use_italics = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "limit_buffer_size" );
	buffer_size = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	if ( buffer_size ) {
		tmp_widget = lookup_widget( setup_window, "buffer_size" );
		buffer_size = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON( tmp_widget ));
	}

	tmp_widget = lookup_widget( setup_window, "use_history_limit" );
	history_limit = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	if ( history_limit ) {
		tmp_widget = lookup_widget( setup_window, "history_limit" );
		history_limit = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON( tmp_widget ));
	}

	tmp_widget = lookup_widget( setup_window, "use_ping_time" );
	ping_minutes = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	if ( ping_minutes ) {
		tmp_widget = lookup_widget( setup_window, "ping_minutes" );
		ping_minutes = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON( tmp_widget ));
	}

	tmp_widget = lookup_widget( setup_window, "force_lowercase" );
	force_lowercase = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "suppress_mult" );
	suppress_mult = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "ignore_guests" );
	ignore_guests = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "auto_reconnect" );
	auto_reconnect = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "show_colors" );
	show_colors = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "show_fonts" );
	show_fonts = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "show_html" );
	show_html = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "show_statuses" );
	show_statuses = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "show_enters" );
	show_enters = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "profile_viewer" );
	profile_viewer = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "ignore_on_mults" );
	ignore_on_mults = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "pm_in_sep_windows" );
	pm_in_sep_windows = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "fontselection1" );
	free( font_name );
	font_name = strdup( gtk_font_selection_get_font_name(
		GTK_FONT_SELECTION(tmp_widget)));
	ptr = font_name;
	ptr = strchr( ptr, '-' ); ptr++;
	ptr = strchr( ptr, '-' ); ptr++;
	free( font_family );
	font_family = strdup( ptr );
	ptr = strchr( font_family, '-' );
	*ptr = '\0';
	ptr++;
	ptr = strchr( ptr, '-' ); ptr++;
	ptr = strchr( ptr, '-' ); ptr++;
	ptr = strchr( ptr, '-' ); ptr++;
	ptr = strchr( ptr, '-' ); ptr++;
	font_size = atoi( ptr );

	/* get the user's color */
	free( use_color );
	use_color = NULL;
	if (( tmp_widget = lookup_widget( setup_window, "color_black" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "black" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_red" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "red" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_blue" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "blue" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_green" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "green" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_yellow" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "yellow" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_purple" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "purple" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_cyan" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "cyan" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_orange" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "orange" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_gray" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "gray" );
	}

	if ( ! use_color )
		use_color = strdup( "black" );

	auto_away_time = time(NULL) + ( auto_away * 60 );

	/* now get and save the friends, regex, and ignore lists */
	tmp_widget = lookup_widget( setup_window, "login_list" );
	/* empty current login list */
	while( login_list ) {
		free( login_list->data );
		list_ptr = login_list;
		login_list = g_list_remove_link( login_list, list_ptr );
		g_list_free( list_ptr );
	}

	/* fetch new login list */
	i = 0;
	while(1) {
		if ( !gtk_clist_get_text(GTK_CLIST(tmp_widget), i, 0, &tmp_str )) {
			break;
		}

		login_list = g_list_append( login_list, strdup( tmp_str ));

		i++;
	}

	tmp_widget = lookup_widget( setup_window, "favroom_list" );
	/* empty current favroom list */
	while( favroom_list ) {
		free( favroom_list->data );
		list_ptr = favroom_list;
		favroom_list = g_list_remove_link( favroom_list, list_ptr );
		g_list_free( list_ptr );
	}

	/* fetch new favroom list */
	i = 0;
	while(1) {
		if ( !gtk_clist_get_text(GTK_CLIST(tmp_widget), i, 0, &tmp_str )) {
			break;
		}

		favroom_list = g_list_append( favroom_list, strdup( tmp_str ));

		i++;
	}


	tmp_widget = lookup_widget( setup_window, "regex_list" );
	sprintf( tmp_filename, "%s/.gyach/ignore.regex", getenv( "HOME" ));
	fp = fopen( tmp_filename, "wb" );

	if ( fp ) {
		i = 0;
		while(1) {
			if ( !gtk_clist_get_text(GTK_CLIST(tmp_widget), i, 0, &tmp_str )) {
				break;
			}

			fprintf( fp, "%s\n", tmp_str );

			i++;
		}
		fclose( fp );
	}

	tmp_widget = lookup_widget( setup_window, "ignore_list" );
	sprintf( tmp_filename, "%s/.gyach/ignore.list", getenv( "HOME" ));
	fp = fopen( tmp_filename, "wb" );

	if ( fp ) {
		i = 0;
		while(1) {
			if ( !gtk_clist_get_text(GTK_CLIST(tmp_widget), i, 0, &tmp_str )) {
				break;
			}

			fprintf( fp, "%s\n", tmp_str );

			i++;
		}
		fclose( fp );
	}

	/* now load the files we just saved */
	ignore_load();

	/* set the options on the toolbar */
	set_config();
}

/* fetch a URL into a predefined buffer (max size 64K) */
int fetch_url( char *image_url, char *buf, char *cookie ) {
	int sock;
	struct sockaddr_in sa;
	struct hostent *hinfo;
	char url[1025];
	static char result[65535];
	char tmp[512];
	char *ptr;
	int nr = 0, tr = 0;
	char host[75] = "";

	if ( strncmp( image_url, "http://", 7 ))
		return( 0 );

	result[0] = '\0';

	strncpy( host, image_url + 7, 74 );
	host[74] = '\0';
	ptr = strchr( host, '/' );
	*ptr = '\0';

	ptr = strchr( image_url + 7, '/' );

	hinfo = gethostbyname( host );

	if ( ! hinfo || ( hinfo->h_addrtype != AF_INET )) {
		/* fixme, show an error box */
		return( 0 );
	}

	memset( &sa, 0, sizeof(sa));
	memmove((void*)&sa.sin_addr.s_addr, hinfo->h_addr, hinfo->h_length );

	sa.sin_family = AF_INET;
	sa.sin_port = htons( 80 );

	if (( sock = socket( AF_INET, SOCK_STREAM, 6 )) == -1 ) {
		/* fixme, error message */
		if ( DBG_LEVEL > 10 ) {
			printf( "error in socket()\n" );
		}
		return( 0 );
	} else {
		if ( connect( sock, (struct sockaddr*)&sa, sizeof(sa)) == -1 ) {
			/* fixme, error message */
			if ( DBG_LEVEL > 10 ) {
				printf( "error in connect()\n" );
			}
			close(sock);
			return( 0 );
		}
	}

	snprintf( url, 1024,
		"GET %s HTTP/1.0\r\n"
		"Accept: text/html\r\n"
		"Accept: text/plain\r\n"
		"User-Agent: Gyach/%s\r\n",
		ptr, VERSION );

	if ( cookie ) {
		strcat( url, "Cookie: " );
		strcat( url, cookie );
		strcat( url, "\r\n" );
	}

	strcat( url, "\r\n\r\n" );

	write( sock, url, strlen( url ));

	result[0] = '\0';

	tr = 0;
	nr = read( sock, tmp, 512 );
	while(( nr > 0 ) &&
		( tr < 65535 )) {
		process_gtk_events();
		tmp[nr] = '\0';
		memcpy( result + tr, tmp, nr );
		tr += nr;

		nr = read( sock, tmp, 512 );
	}

	close( sock );

	result[tr] = '\0';

	ptr = strstr( result, "\r\n\r\n" );
	ptr += 4;

	strcpy( buf, result );

	return( strlen( result ));
}

/* this function and populate_room_list() make up the 2nd time I've      */
/* written a function to parse yahoo's roomlist.html page, let's see how */
/* long they keep it the same before they change it again.  Oct 10, 2000 */
void populate_child_room_list( char *number, GtkCTree *room_tree,
	GtkCTreeNode *parent_node )
{
	GtkCTreeNode *child_node;
	GtkLabel *room_label;
	ROOM_ENTRY *room_item;
	char url[129];
	char buf[65536];
	char name[129];
	char *data[1];
	char *ptr;
	char *end;
	int url_length;

	DBG( 11, "populate_child_room_list()\n" );

	room_label =
		GTK_LABEL( lookup_widget( GTK_WIDGET(room_tree), "loading_status" ));

	data[0] = name;

	sprintf( url, "http://chat.yahoo.com/c/roomlist.html?.rmcat=%s", number );

	gtk_label_set_text( room_label, "Retrieving Room List..." );
	url_length = fetch_url( url, buf, GlobalSessionHook->cook );
	gtk_label_set_text( room_label, "Parsing Room List..." );

	ptr = strstr( buf, ">Yahoo! Rooms<" );
	if ( ! ptr ) {
		return;
	}

	process_gtk_events();

	ptr++;
	ptr = strstr( ptr, "href=\"/?room=" );
	while(( ptr ) &&
		( *ptr )) {
		ptr = strchr( ptr, '>' );
		ptr++;

		strncpy( name, ptr, sizeof( name ) - 1 );
		end = strchr( name, '<' );
		*end = '\0';

		/* insert the child */
		if ( strlen( name )) {
			room_item = (ROOM_ENTRY *)malloc( sizeof( ROOM_ENTRY ));
			room_item->room_name = strdup( name );
			room_item->room_id = NULL;
			room_item->populated = 0;

			child_node = gtk_ctree_insert_node( room_tree, parent_node,
							NULL, data, 5, NULL, NULL, NULL, NULL,
							FALSE, FALSE);

			/* last field is 1 = selectable, 0 = not */
			gtk_ctree_node_set_selectable( room_tree, child_node, 1 );
			gtk_ctree_node_set_row_data( room_tree, child_node, room_item );

			process_gtk_events();
		}

		ptr = strstr( ptr, "href=\"/?room=" );
	}

	gtk_label_set_text( room_label, "Room List" );
}


void populate_room_list( void ) {
	GtkCTree *room_tree;
	GtkCTreeNode *nodes[21];
	GtkLabel *room_label;
	ROOM_ENTRY *room_item;
	char title[129];
	char buf[65536];
	char name[129];
	char number[31];
	char *data[1];
	char *ptr;
	char *end;
	char *space;
	int level = 1;
	int url_length;

	DBG( 11, "populate_room_list()\n" );

	data[0] = name;

	room_tree = GTK_CTREE(lookup_widget( room_window, "room_tree" ));

	room_label =
		GTK_LABEL( lookup_widget( GTK_WIDGET(room_tree), "loading_status" ));

	gtk_label_set_text( room_label, "Retrieving Room List..." );

	process_gtk_events();

	url_length = fetch_url( "http://chat.yahoo.com/c/roomlist.html", buf,
		GlobalSessionHook->cook );

	gtk_label_set_text( room_label, "Loading Room List..." );
	gtk_ctree_set_line_style( room_tree, GTK_CTREE_LINES_NONE);
	gtk_ctree_set_expander_style( room_tree, GTK_CTREE_EXPANDER_TRIANGLE);

	process_gtk_events();

	nodes[0] = NULL;

	gtk_clist_freeze( GTK_CLIST(room_tree) );

	ptr = strstr( buf, "<table" );
	if ( ! ptr ) {
		return;
	}

	ptr = strstr( ptr + 1, "<table" );
	if ( ! ptr ) {
		return;
	}

	ptr = strstr( ptr + 1, "<tr" );
	if ( ! ptr ) {
		return;
	}

	ptr++;
	ptr = strstr( ptr, "href=\"/c/roomlist.html" );
	while(( ptr ) &&
		( *ptr )) {
		space = ptr - 40;
		space = strstr( space, "&nbsp;" );
		level = 1;
		space = strstr( space + 1, "&nbsp;&nbsp;" );
		while(( space ) && ( space < ptr )) {
			level++;
			space = strstr( space + 12, "&nbsp;&nbsp;" );
		}

		ptr = strstr( ptr + 1, "name=" );
		if ( ! ptr ) {
			continue;
		}

		ptr += 5;
		strncpy( number, ptr, sizeof( number ) - 1 );
		end = strchr( number, '>' );
		*end = '\0';

		ptr = strchr( ptr, '>' );
		ptr++;

		strncpy( name, ptr, sizeof( name ) - 1 );
		end = strchr( name, '<' );
		*end = '\0';

		/* now insert the item into the tree */
		if ( strlen( name )) {
			/* check the parent to see if we need to remove temp child */
			room_item = gtk_ctree_node_get_row_data( room_tree, nodes[level-1]);
			if (( room_item ) &&
				( room_item->temp_child )) {
				gtk_ctree_remove_node( room_tree, room_item->temp_child );
				room_item->populated = 1;
				room_item->temp_child = NULL;
			}


			room_item = (ROOM_ENTRY *)malloc( sizeof( ROOM_ENTRY ));
			room_item->room_name = strdup( name );
			room_item->room_id = strdup( number );
			room_item->populated = 0;
			room_item->temp_child = NULL;

			nodes[level] = gtk_ctree_insert_node( room_tree, nodes[level-1],
							NULL, data, 5, NULL, NULL, NULL, NULL,
							FALSE, FALSE);

			/* last field is 1 = selectable, 0 = not */
			gtk_ctree_node_set_selectable( room_tree, nodes[level], 0 );
			gtk_ctree_node_set_row_data( room_tree, nodes[level], room_item );

			/* now download and insert the children */
			sprintf( title, "Retrieving %s List...", name );
			gtk_label_set_text( room_label, title );
			gtk_clist_thaw( GTK_CLIST(room_tree) );
			process_gtk_events();
			gtk_clist_freeze( GTK_CLIST(room_tree) );

			strcpy( name, "---- downloading list ----" );
			room_item->temp_child = gtk_ctree_insert_node( room_tree,
					nodes[level], NULL, data, 5,
						NULL, NULL, NULL, NULL, FALSE, FALSE);
			gtk_ctree_node_set_selectable( room_tree,
				room_item->temp_child, 0 );
		}

		ptr = strstr( ptr, "href=\"/c/roomlist.html" );
	}

	gtk_label_set_text( room_label, "Room List" );
	/* this will hide the column header at the top of the room list.
	gtk_clist_column_titles_hide( GTK_CLIST(room_tree));
	*/

	gtk_clist_thaw( GTK_CLIST(room_tree) );
}

void subst_escs( char *str ) {
	char tmp[2048];
	char *from = tmp;
	char *to = str;

	DBG( 21, "subst_escs( '%s' )\n", str );

	if ( ! strchr( str, '\\' ))
		return;

	strcpy( tmp, str );

	while( *from ) {
		if (( *from == '\\' ) &&
			( *(from+1) == 'n' )) {
			*(to++) = '\n';
			from += 2;
		} else {
			*(to++) = *(from++);
		}
	}
	*to = '\0';
}

void history_add( char *str ) {
	GList *this_link;
	static char last[129] = "";

	DBG( 11, "history_add( '%s' )\n", str );

	history_current = history_size;

	/* don't add the same entry twice in a row */
	if ( ! strcmp( last, str ))
		return;

	strcpy( last, str );

	history_size++;

	while ( history_size > ( history_limit + 1 )) {
		history_size--;
		this_link = history_list;
		history_list = g_list_remove_link( history_list, this_link );
		free( this_link->data );
		g_list_free( this_link );
		strcpy( history_list->data, "" );
	}

	history_current = history_size;
	history_list = g_list_append( history_list, strdup( str ));
}

char *history_next() {
	GList *this_item;

	DBG( 11, "history_next()\n" );

	if ( ! history_size )
		return( "" );

	history_current++;

	if ( history_current >= history_size )
		history_current = 0;

	this_item = g_list_nth( history_list, history_current );

	return( this_item->data );
}

char *history_prev() {
	GList *this_item;

	DBG( 11, "history_prev()\n" );

	if ( ! history_size )
		return( "" );

	history_current--;

	if ( history_current < 0 )
		history_current = history_size - 1;

	this_item = g_list_nth( history_list, history_current );

	return( this_item->data );
}

void history_print() {
	GList *this_item;
	char buf[129];

	DBG( 11, "history_print()\n" );

	append_to_textbox( chat_window, chat_text, "*** Command History ***\n" );

	this_item = history_list;
	while( this_item ) {
		if ( strcmp( this_item->data, "" )) {
			sprintf( buf, "%s\n", (char *)this_item->data );
			append_to_textbox( chat_window, chat_text, buf );
		}
		this_item = g_list_next( this_item );
	}
	append_to_textbox( chat_window, chat_text, "*** End History ***\n" );
}

char *filter_text( char *txt ) {
	static char buf[513] = "";
	char data[128];
	char *ptr;
	char new_cmd[513];
	FILE *fp;

	DBG( 11, "filter_text( '%s' )\n", txt );

	/* zero out buf every time */
	buf[0] = '\0';

	if ( ! filter_command ) {
		strcpy( buf, txt );
		return( buf );
	}

	/* sanity check to make sure our echo doesn't hose up */
	ptr = txt;
	while( *ptr ) {
		if ( *ptr == '"' )
			*ptr = '\'';
		ptr++;
	}
	sprintf( new_cmd, "echo \"%s\" | %s", txt, filter_command );

	fp = popen( new_cmd, "r" );

	if ( fp ) {
		while(( strlen( buf ) < 512 ) &&
			  ( fgets( data, 512-strlen(buf), fp ))) {
			if (( strlen( buf ) + strlen( data )) < 512 )
				strcat( buf, data );
		}
		pclose( fp );

		ptr = buf + strlen( buf );
		ptr--;
		while(( *ptr == '\n' ) || ( *ptr == '\r' )) {
			*ptr = '\0';
			ptr--;
		}

		if ( strlen( buf ) == 511 ) {
			strcpy( buf+507, "...." );
		}
	} else {
		strcpy( buf, txt );
	}

	return( buf );
}

GList *find_pm_session( char *user ) {
	GList *this_session = pm_list;
	PM_SESSION *pm_sess;

	DBG( 11, "find_pm_session( '%s' )\n", user );

	while( this_session ) {
		pm_sess = (PM_SESSION *)this_session->data;
		if ( !strcasecmp( user, pm_sess->pm_user )) {
			return( this_session );
		}

		this_session = g_list_next( this_session );
	}

	return( NULL );
}

GList *find_pm_session_from_widget( GtkWidget *widget, char *name ) {
	GList *this_session = pm_list;
	PM_SESSION *pm_sess;
	GtkWidget *tmp_widget;

	DBG( 11, "find_pm_session_from_widget( %p, '%s' )\n", widget, name );

	while( this_session ) {
		pm_sess = (PM_SESSION *)this_session->data;

		if ( pm_sess->pm_window == widget ) {
			return( this_session );
		}

		tmp_widget = lookup_widget( pm_sess->pm_window, name );
		if ( tmp_widget == widget ) {
			return( this_session );
		}

		this_session = g_list_next( this_session );
	}

	return( NULL );
}

char *replace_args( char *str, char *args ) {
	static char result[1025] = "";
	char zargs[1025];
	char tmp[1025];
	char search[3];
	char *arg;
	char *ptr;
	char *end;
	int x;

	DBG( 11, "replace_args( '%s', '%s' )\n", str, args );

	strcpy( result, str );
	strcpy( zargs, args );

	if ( ! strchr( result, '$' ))
		return( result );

	arg = zargs;
	for( x = 0; x < 10; x++ ) {
		/* find next arg */
		arg = skip_whitespace( arg );
		end = find_whitespace( arg );
		*end = '\0';

		sprintf( search, "$%d", x );
		while(( ptr = strstr( result, search )) != NULL ) {
			strcpy( tmp, ptr + 2 );

			/* append arg if we have room */
			if (( ptr - result + strlen( arg ) + 1 ) < sizeof( result )) 
				strcpy( ptr, arg );

			/* append rest of string after arg if we have room */
			strncat( ptr, tmp, sizeof( result ) - strlen( result ));
		}

		arg = end + 1;
	}

	while(( ptr = strstr( result, "$*" )) != NULL ) {
		strcpy( tmp, ptr + 2 );

		/* append arg if we have room */
		if (( ptr - result + strlen( args ) + 1 ) < sizeof( result )) 
			strcpy( ptr, args );

		/* append rest of string after arg if we have room */
		strncat( ptr, tmp, sizeof( result ) - strlen( result ));
	}

	return( result );
}

void io_callback(gpointer data, gint source, GdkInputCondition condition)
{
	char buf[1024];

	/* Tell the library to retrieve any newly available data */
	if (condition == GDK_INPUT_READ)
	{
		if ( yahooRecvData( sess )) {

			if(sess->yin.type == 0x01) {
				gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );

				sprintf(buf ,"Chat started, type /help for help\n"); 
				append_to_textbox( chat_window, chat_text, buf );
				YahooTextCommand(sess ,"join " ,sess->room); 

				sprintf( buf, "Current Room: %s", sess->room );
				gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid, buf );

				/* set the return value for the program */ 
				sess->rval = 0; 
			} 

			showYahooPacket( sess ); 

			if (( sess->yin.type == 0x02 ) &&
				( auto_reconnect )) {
				on_disconnect_activate( NULL, NULL );
				login_to_yahoo_chat();
			}
		}
	} else {
		/* invalid/unknown condition, so disconnect */
		on_disconnect_activate( NULL, NULL );
	}
}

int ping_callback( guint data ) {
	if ( logged_in )
		yahooPing( sess );

	return( 1 );
}

int login_to_yahoo_chat( void ) {
#ifdef USE_PTHREAD_CREATE
	pthread_t friend_thread;
#endif


	if ( LoginToYahoo( sess )) {
		if ( LoginToChat( sess )) {
			gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
			gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
				"Waiting for Yahoo! greeting..." );

			logged_in = 1;

			yahoo_ctx_init();

#ifdef USE_PTHREAD_CREATE
			pthread_create( &friend_thread, NULL, fetch_friend_list, NULL );
#else
			fetch_friend_list( NULL );
#endif

			sleep( 1 );

			io_callback_tag = gdk_input_add( sess->sock, GDK_INPUT_READ,
				io_callback, (gpointer) 0);

			ping_callback_tag = gtk_timeout_add( ping_minutes * 60 * 1000,
				(void *)ping_callback, NULL );

			set_menu_connected( 1 );
		} else {
			gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
			gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
				"Error logging in to Yahoo! Chat..." );
		}
	} else {
		gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
		gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
			"Error logging in to Yahoo!" );
	}

	return( 0 );
}

void *fetch_friend_list( void *arg ) {
	GList *tmp_list = NULL;
	struct yahoo_buddy *buddy;
	int x;

	if ( friend_list ) {
		while( friend_list ) {
			tmp_list = friend_list;
			friend_list = g_list_remove_link( friend_list, friend_list );
			free( tmp_list->data );
			g_list_free( tmp_list );
		}
	}

	yahoo_get_config( &yahoo_ctx );
	x = 0;
	tmp_list = NULL;
	if ( yahoo_ctx.buddies ) {
		buddy = yahoo_ctx.buddies[x];
		while( buddy ) {
			tmp_list = g_list_append( tmp_list, strdup( buddy->id ));
			x++;
			buddy = yahoo_ctx.buddies[x];
		}

		/* now set the real list */
		friend_list = tmp_list;
	}

	return( NULL );
}

int remove_friend( char *friend ) {
	GList *this_friend;

	yahoo_remove_buddy( &yahoo_ctx, friend, sess->user, "friends", "" );

	this_friend = friend_list;
	while( this_friend ) {
		if ( ! strcasecmp( this_friend->data, friend )) {
			friend_list = g_list_remove_link( friend_list, this_friend );
			free( this_friend->data );
			g_list_free( this_friend );
			return( 0 );
		}
		this_friend = g_list_next( this_friend );
	}

	return( 0 );
}

int add_friend( char *friend ) {
	yahoo_add_buddy( &yahoo_ctx, friend, sess->user, "friends", "" );

	friend_list = g_list_append( friend_list, strdup( friend ));

	return( 0 );
}

int find_friend( char *friend ) {
	GList *this_friend;

	this_friend = friend_list;
	while( this_friend ) {
		if ( ! strcasecmp( this_friend->data, friend )) {
			return( 1 );
		}
		this_friend = g_list_next( this_friend );
	}

	return( 0 );
}

int remove_online_friend( char *friend ) {
	GList *this_friend;

	this_friend = ofriend_list;
	while( this_friend ) {
		if ( ! strcasecmp( this_friend->data, friend )) {
			ofriend_list = g_list_remove_link( ofriend_list, this_friend );
			free( this_friend->data );
			g_list_free( this_friend );
			return( 0 );
		}
		this_friend = g_list_next( this_friend );
	}

	return( 0 );
}

int add_online_friend( char *friend ) {
	ofriend_list = g_list_append( ofriend_list, strdup( friend ));

	return( 0 );
}

int find_online_friend( char *friend ) {
	GList *this_friend;

	this_friend = ofriend_list;
	while( this_friend ) {
		if ( ! strcasecmp( this_friend->data, friend )) {
			return( 1 );
		}
		this_friend = g_list_next( this_friend );
	}

	return( 0 );
}

int show_friends( void ) {
	GList *this_friend;
	char msg[4096];
	int any_online = 0;


	if ( ! friend_list )
		return( 0 );

	sprintf( msg, "%s%sCurrent Buddylist:%s%s ", YAHOO_COLOR_BLUE,
		YAHOO_STYLE_BOLDON, YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF );

	this_friend = friend_list;
	while( this_friend ) {
		if ( this_friend != friend_list )
			strcat( msg, ", " );

		if ( find_online_friend( this_friend->data )) {
			strcat( msg, YAHOO_COLOR_RED );
			strcat( msg, this_friend->data );
			strcat( msg, YAHOO_COLOR_BLACK );
			any_online = 1;
		} else {
			strcat( msg, this_friend->data );
		}

		this_friend = g_list_next( this_friend );
	}

	if ( ! friend_list ) {
		strcat( msg, "No known friends." );
	} else {
		if ( any_online ) {
			strcat( msg, "...  (friends online in " );
				strcat( msg, YAHOO_COLOR_RED );
				strcat( msg, "RED" );
				strcat( msg, YAHOO_COLOR_BLACK );
				strcat( msg, ")" );
		}
	}

	strcat( msg, "\n" );

	append_to_textbox_color( chat_window, chat_text, msg );

	return( 0 );
}

int build_online_friends_list( void ) {
	GList *tmp_list = NULL;
	char *ptr;
	char *end;

	if ( ofriend_list ) {
		while( ofriend_list ) {
			tmp_list = ofriend_list;
			ofriend_list = g_list_remove_link( ofriend_list, ofriend_list );
			free( tmp_list->data );
			g_list_free( tmp_list );
		}
	}

	sess->yin.data[sess->yin.size] = '\0';
	ptr = sess->yin.data;

	/* find first field separator */
	while(( ptr < ( sess->yin.data + sess->yin.size )) &&
		( *ptr != '\0' ) &&
		( *ptr != '\300' )) {
		ptr++;
	}
	ptr += 2;

	/* now find the names */
	while( ptr < ( sess->yin.data + sess->yin.size )) {
		end = ptr;
		while( end < ( sess->yin.data + sess->yin.size )) {
			if (( *end == '\300' ) ||
				( *end == '\0' )) {
				*end = '\0';
				break;
			}
			end++;
		}

		ofriend_list = g_list_append( ofriend_list, strdup( ptr ));

		end += 2;
		while( end < ( sess->yin.data + sess->yin.size )) {
			if (( *end == '\300' ) ||
				( *end == '\0' )) {
				*end = '\0';
				break;
			}
			end++;
		}
		ptr = end + 2;
	}

	return( 0 );
}

static char *conn_menu_items[] = {
		"disconnect",	"save_current_text",	"log_to_file",
		"favorites",	"add_to_favorites", 	"rooms",
		"status",
		NULL
	};

static char *disconn_menu_items[] = {
		"connect",
		NULL
	};

int set_menu_connected( int connected ) {
	GtkWidget *tmp_widget;
	char **on_items;
	char **off_items;
	int i;

	if ( connected ) {
		on_items = conn_menu_items;
		off_items = disconn_menu_items;
	} else {
		off_items = conn_menu_items;
		on_items = disconn_menu_items;
	}

	i = 0;
	while( on_items[i] ) {
		tmp_widget = lookup_widget( chat_window, on_items[i] );
		gtk_widget_set_sensitive( tmp_widget, 1 );
		i++;
	}

	i = 0;
	while( off_items[i] ) {
		tmp_widget = lookup_widget( chat_window, off_items[i] );
		gtk_widget_set_sensitive( tmp_widget, 0 );
		i++;
	}

	return( 0 );
}
